# -*- coding: utf-8
# Copyright (C) Dnspython Contributors, see LICENSE for text of ISC license

# Copyright (C) 2003-2017 Nominum, Inc.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose with or without fee is hereby granted,
# provided that the above copyright notice and this permission notice
# appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NOMINUM DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NOMINUM BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import copy
import operator
import pickle
import unittest
from io import BytesIO
from typing import Dict  # pylint: disable=unused-import

import dns.e164
import dns.name
import dns.reversename

# pylint: disable=line-too-long,unsupported-assignment-operation


def expand(text):
    # This is a helper routine to expand name patterns for RFC 4471 tests.
    #
    # Basically it turns <character>{<n>} into <n> instances of the character.
    # For example:
    #
    #       r"fo{2}.example." => r"foo.example.".
    #
    # Two characters get special treatment:
    # "-" is mapped to r"\000" and "+" is mapped to r"\255".  For example
    #
    #       r"+{3}-.example." -> r"\255\255\255\000.example."
    #
    # We do this just to make parsing simpler, so we don't have to process escapes
    # ourselves.
    i = 0
    l = len(text)
    previous = ""
    reading_count = False
    count = 0
    expanded = []
    for c in text:
        if c == "-":
            c = r"\000"
        elif c == "+":
            c = r"\255"
        if reading_count:
            assert len(c) == 1
            if c >= "0" and c <= "9":
                count *= 10
                count += ord(c) - ord("0")
            elif c == "}":
                expanded.append(previous * count)
                previous = ""
                reading_count = False
                count = 0
        elif c == "{":
            reading_count = True
        else:
            expanded.append(previous)
            previous = c
    # don't forget the last char (if there is one)
    expanded.append(previous)
    x = "".join(expanded)
    return x


class NameTestCase(unittest.TestCase):
    def setUp(self):
        self.origin = dns.name.from_text("example.")

    def testFromTextRel1(self):
        n = dns.name.from_text("foo.bar")
        self.assertEqual(n.labels, (b"foo", b"bar", b""))

    def testFromTextRel2(self):
        n = dns.name.from_text("foo.bar", origin=self.origin)
        self.assertEqual(n.labels, (b"foo", b"bar", b"example", b""))

    def testFromTextRel3(self):
        n = dns.name.from_text("foo.bar", origin=None)
        self.assertEqual(n.labels, (b"foo", b"bar"))

    def testFromTextRel4(self):
        n = dns.name.from_text("@", origin=None)
        self.assertEqual(n, dns.name.empty)

    def testFromTextRel5(self):
        n = dns.name.from_text("@", origin=self.origin)
        self.assertEqual(n, self.origin)

    def testFromTextAbs1(self):
        n = dns.name.from_text("foo.bar.")
        self.assertEqual(n.labels, (b"foo", b"bar", b""))

    def testTortureFromText(self):
        good = [
            rb".",
            rb"a",
            rb"a.",
            rb"a" * 63,
            (rb"a" * 63 + rb".") * 3 + rb"a" * 61,
            rb"\000.\008.\010.\032.\046.\092.\099.\255",
            rb"\\",
            rb"\..\.",
            rb"\\.\\",
            rb'!"#%&/()=+-',
            (rb"\255" * 63 + rb".") * 3 + rb"\255" * 61,
        ]
        bad = [
            rb"..",
            rb".a",
            rb"\\..",
            b"\\",  # yes, we don't want the 'r' prefix!
            rb"\0",
            rb"\00",
            rb"\00Z",
            rb"a" * 64,
            (rb"a" * 63 + rb".") * 3 + rb"a" * 62,
            (rb"\255" * 63 + rb".") * 3 + rb"\255" * 62,
        ]
        for t in good:
            try:
                dns.name.from_text(t)
            except Exception:
                self.fail("good test '%r' raised an exception" % t)
        for t in bad:
            caught = False
            try:
                dns.name.from_text(t)
            except Exception:
                caught = True
            if not caught:
                self.fail("bad test '%r' did not raise an exception" % t)

    def testImmutable1(self):
        def bad():
            self.origin.labels = ()

        self.assertRaises(TypeError, bad)

    def testImmutable2(self):
        def bad():
            self.origin.labels[0] = "foo"  # type: ignore

        self.assertRaises(TypeError, bad)

    def testAbs1(self):
        self.assertTrue(dns.name.root.is_absolute())

    def testAbs2(self):
        self.assertFalse(dns.name.empty.is_absolute())

    def testAbs3(self):
        self.assertTrue(self.origin.is_absolute())

    def testAbs4(self):
        n = dns.name.from_text("foo", origin=None)
        self.assertFalse(n.is_absolute())

    def testWild1(self):
        n = dns.name.from_text("*.foo", origin=None)
        self.assertTrue(n.is_wild())

    def testWild2(self):
        n = dns.name.from_text("*a.foo", origin=None)
        self.assertFalse(n.is_wild())

    def testWild3(self):
        n = dns.name.from_text("a.*.foo", origin=None)
        self.assertFalse(n.is_wild())

    def testWild4(self):
        self.assertFalse(dns.name.root.is_wild())

    def testWild5(self):
        self.assertFalse(dns.name.empty.is_wild())

    def testHash1(self):
        n1 = dns.name.from_text("fOo.COM")
        n2 = dns.name.from_text("foo.com")
        self.assertEqual(hash(n1), hash(n2))

    def testCompare1(self):
        n1 = dns.name.from_text("a")
        n2 = dns.name.from_text("b")
        self.assertLess(n1, n2)
        self.assertLessEqual(n1, n2)
        self.assertGreater(n2, n1)
        self.assertGreaterEqual(n2, n1)

    def testCompare2(self):
        n1 = dns.name.from_text("")
        n2 = dns.name.from_text("b")
        self.assertLess(n1, n2)
        self.assertLessEqual(n1, n2)
        self.assertGreater(n2, n1)
        self.assertGreaterEqual(n2, n1)

    def testCompare3(self):
        self.assertLess(dns.name.empty, dns.name.root)
        self.assertGreater(dns.name.root, dns.name.empty)

    def testCompare4(self):
        self.assertNotEqual(dns.name.root, 1)

    def testSubdomain1(self):
        self.assertFalse(dns.name.empty.is_subdomain(dns.name.root))

    def testSubdomain2(self):
        self.assertFalse(dns.name.root.is_subdomain(dns.name.empty))

    def testSubdomain3(self):
        n = dns.name.from_text("foo", origin=self.origin)
        self.assertTrue(n.is_subdomain(self.origin))

    def testSubdomain4(self):
        n = dns.name.from_text("foo", origin=self.origin)
        self.assertTrue(n.is_subdomain(dns.name.root))

    def testSubdomain5(self):
        n = dns.name.from_text("foo", origin=self.origin)
        self.assertTrue(n.is_subdomain(n))

    def testSuperdomain1(self):
        self.assertFalse(dns.name.empty.is_superdomain(dns.name.root))

    def testSuperdomain2(self):
        self.assertFalse(dns.name.root.is_superdomain(dns.name.empty))

    def testSuperdomain3(self):
        n = dns.name.from_text("foo", origin=self.origin)
        self.assertTrue(self.origin.is_superdomain(n))

    def testSuperdomain4(self):
        n = dns.name.from_text("foo", origin=self.origin)
        self.assertTrue(dns.name.root.is_superdomain(n))

    def testSuperdomain5(self):
        n = dns.name.from_text("foo", origin=self.origin)
        self.assertTrue(n.is_superdomain(n))

    def testCanonicalize1(self):
        n = dns.name.from_text("FOO.bar", origin=self.origin)
        c = n.canonicalize()
        self.assertEqual(c.labels, (b"foo", b"bar", b"example", b""))

    def testToText1(self):
        n = dns.name.from_text("FOO.bar", origin=self.origin)
        t = n.to_text()
        self.assertEqual(t, "FOO.bar.example.")

    def testToText2(self):
        n = dns.name.from_text("FOO.bar", origin=self.origin)
        t = n.to_text(True)
        self.assertEqual(t, "FOO.bar.example")

    def testToText3(self):
        n = dns.name.from_text("FOO.bar", origin=None)
        t = n.to_text()
        self.assertEqual(t, "FOO.bar")

    def testToText4(self):
        t = dns.name.empty.to_text()
        self.assertEqual(t, "@")

    def testToText5(self):
        t = dns.name.root.to_text()
        self.assertEqual(t, ".")

    def testToText6(self):
        n = dns.name.from_text("FOO bar", origin=None)
        t = n.to_text()
        self.assertEqual(t, r"FOO\032bar")

    def testToText7(self):
        n = dns.name.from_text(r"FOO\.bar", origin=None)
        t = n.to_text()
        self.assertEqual(t, r"FOO\.bar")

    def testToText8(self):
        n = dns.name.from_text(r"\070OO\.bar", origin=None)
        t = n.to_text()
        self.assertEqual(t, r"FOO\.bar")

    def testToText9(self):
        n = dns.name.from_text("FOO bar", origin=None)
        t = n.to_unicode()
        self.assertEqual(t, "FOO\\032bar")

    def testToText10(self):
        t = dns.name.empty.to_unicode()
        self.assertEqual(t, "@")

    def testToText11(self):
        t = dns.name.root.to_unicode()
        self.assertEqual(t, ".")

    def testToText12(self):
        n = dns.name.from_text(r"a\.b.c")
        t = n.to_unicode()
        self.assertEqual(t, r"a\.b.c.")

    def testToText13(self):
        n = dns.name.from_text(r"\150\151\152\153\154\155\156\157\158\159.")
        t = n.to_text()
        self.assertEqual(t, r"\150\151\152\153\154\155\156\157\158\159.")

    def testToText14(self):
        # Something that didn't start as unicode should go to escapes and not
        # raise due to interpreting arbitrary binary DNS labels as UTF-8.
        n = dns.name.from_text(r"\150\151\152\153\154\155\156\157\158\159.")
        t = n.to_unicode()
        self.assertEqual(t, r"\150\151\152\153\154\155\156\157\158\159.")

    def testSlice1(self):
        n = dns.name.from_text(r"a.b.c.", origin=None)
        s = n[:]
        self.assertEqual(s, (b"a", b"b", b"c", b""))

    def testSlice2(self):
        n = dns.name.from_text(r"a.b.c.", origin=None)
        s = n[:2]
        self.assertEqual(s, (b"a", b"b"))

    def testSlice3(self):
        n = dns.name.from_text(r"a.b.c.", origin=None)
        s = n[2:]
        self.assertEqual(s, (b"c", b""))

    def testEmptyLabel1(self):
        def bad():
            dns.name.Name(["a", "", "b"])

        self.assertRaises(dns.name.EmptyLabel, bad)

    def testEmptyLabel2(self):
        def bad():
            dns.name.Name(["", "b"])

        self.assertRaises(dns.name.EmptyLabel, bad)

    def testEmptyLabel3(self):
        n = dns.name.Name(["b", ""])
        self.assertTrue(n)

    def testLongLabel(self):
        n = dns.name.Name(["a" * 63])
        self.assertTrue(n)

    def testLabelTooLong(self):
        def bad():
            dns.name.Name(["a" * 64, "b"])

        self.assertRaises(dns.name.LabelTooLong, bad)

    def testLongName(self):
        n = dns.name.Name(["a" * 63, "a" * 63, "a" * 63, "a" * 62])
        self.assertTrue(n)

    def testNameTooLong(self):
        def bad():
            dns.name.Name(["a" * 63, "a" * 63, "a" * 63, "a" * 63])

        self.assertRaises(dns.name.NameTooLong, bad)

    def testConcat1(self):
        n1 = dns.name.Name(["a", "b"])
        n2 = dns.name.Name(["c", "d"])
        e = dns.name.Name(["a", "b", "c", "d"])
        r = n1 + n2
        self.assertEqual(r, e)

    def testConcat2(self):
        n1 = dns.name.Name(["a", "b"])
        n2 = dns.name.Name([])
        e = dns.name.Name(["a", "b"])
        r = n1 + n2
        self.assertEqual(r, e)

    def testConcat3(self):
        n1 = dns.name.Name([])
        n2 = dns.name.Name(["a", "b"])
        e = dns.name.Name(["a", "b"])
        r = n1 + n2
        self.assertEqual(r, e)

    def testConcat4(self):
        n1 = dns.name.Name(["a", "b", ""])
        n2 = dns.name.Name([])
        e = dns.name.Name(["a", "b", ""])
        r = n1 + n2
        self.assertEqual(r, e)

    def testConcat5(self):
        n1 = dns.name.Name(["a", "b"])
        n2 = dns.name.Name(["c", ""])
        e = dns.name.Name(["a", "b", "c", ""])
        r = n1 + n2
        self.assertEqual(r, e)

    def testConcat6(self):
        def bad():
            n1 = dns.name.Name(["a", "b", ""])
            n2 = dns.name.Name(["c"])
            return n1 + n2

        self.assertRaises(dns.name.AbsoluteConcatenation, bad)

    def testBadEscape(self):
        def bad():
            n = dns.name.from_text(r"a.b\0q1.c.")

        self.assertRaises(dns.name.BadEscape, bad)

    def testDigestable1(self):
        n = dns.name.from_text("FOO.bar")
        d = n.to_digestable()
        self.assertEqual(d, b"\x03foo\x03bar\x00")

    def testDigestable2(self):
        n1 = dns.name.from_text("FOO.bar")
        n2 = dns.name.from_text("foo.BAR.")
        d1 = n1.to_digestable()
        d2 = n2.to_digestable()
        self.assertEqual(d1, d2)

    def testDigestable3(self):
        d = dns.name.root.to_digestable()
        self.assertEqual(d, b"\x00")

    def testDigestable4(self):
        n = dns.name.from_text("FOO.bar", None)
        d = n.to_digestable(dns.name.root)
        self.assertEqual(d, b"\x03foo\x03bar\x00")

    def testBadDigestable(self):
        def bad():
            n = dns.name.from_text("FOO.bar", None)
            n.to_digestable()

        self.assertRaises(dns.name.NeedAbsoluteNameOrOrigin, bad)

    def testToWire1(self):
        n = dns.name.from_text("FOO.bar")
        f = BytesIO()
        compress = {}  # type: Dict[dns.name.Name,int]
        n.to_wire(f, compress)
        self.assertEqual(f.getvalue(), b"\x03FOO\x03bar\x00")

    def testToWire2(self):
        n = dns.name.from_text("FOO.bar")
        f = BytesIO()
        compress = {}  # type: Dict[dns.name.Name,int]
        n.to_wire(f, compress)
        n.to_wire(f, compress)
        self.assertEqual(f.getvalue(), b"\x03FOO\x03bar\x00\xc0\x00")

    def testToWire3(self):
        n1 = dns.name.from_text("FOO.bar")
        n2 = dns.name.from_text("foo.bar")
        f = BytesIO()
        compress = {}  # type: Dict[dns.name.Name,int]
        n1.to_wire(f, compress)
        n2.to_wire(f, compress)
        self.assertEqual(f.getvalue(), b"\x03FOO\x03bar\x00\xc0\x00")

    def testToWire4(self):
        n1 = dns.name.from_text("FOO.bar")
        n2 = dns.name.from_text("a.foo.bar")
        f = BytesIO()
        compress = {}  # type: Dict[dns.name.Name,int]
        n1.to_wire(f, compress)
        n2.to_wire(f, compress)
        self.assertEqual(f.getvalue(), b"\x03FOO\x03bar\x00\x01\x61\xc0\x00")

    def testToWire5(self):
        n1 = dns.name.from_text("FOO.bar")
        n2 = dns.name.from_text("a.foo.bar")
        f = BytesIO()
        compress = {}  # type: Dict[dns.name.Name,int]
        n1.to_wire(f, compress)
        n2.to_wire(f, None)
        self.assertEqual(f.getvalue(), b"\x03FOO\x03bar\x00\x01\x61\x03foo\x03bar\x00")

    def testToWire6(self):
        n = dns.name.from_text("FOO.bar")
        v = n.to_wire()
        self.assertEqual(v, b"\x03FOO\x03bar\x00")

    def testToWireRelativeNameWithOrigin(self):
        n = dns.name.from_text("FOO", None)
        o = dns.name.from_text("bar")
        v = n.to_wire(origin=o)
        self.assertEqual(v, b"\x03FOO\x03bar\x00")

    def testToWireRelativeNameWithoutOrigin(self):
        n = dns.name.from_text("FOO", None)

        def bad():
            v = n.to_wire()

        self.assertRaises(dns.name.NeedAbsoluteNameOrOrigin, bad)

    def testBadToWire(self):
        def bad():
            n = dns.name.from_text("FOO.bar", None)
            f = BytesIO()
            compress = {}  # type: Dict[dns.name.Name,int]
            n.to_wire(f, compress)

        self.assertRaises(dns.name.NeedAbsoluteNameOrOrigin, bad)

    def testGiantCompressionTable(self):
        # Only the first 16KiB of a message can have compression pointers.
        f = BytesIO()
        compress = {}  # type: Dict[dns.name.Name,int]
        # exactly 16 bytes encoded
        n = dns.name.from_text("0000000000.com.")
        n.to_wire(f, compress)
        # There are now two entries in the compression table (for the full
        # name, and for the com. suffix.
        self.assertEqual(len(compress), 2)
        for i in range(1023):
            # exactly 16 bytes encoded with compression
            n = dns.name.from_text(f"{i:013d}.com")
            n.to_wire(f, compress)
        # There are now 1025 entries in the compression table with
        # the last entry at offset 16368.
        self.assertEqual(len(compress), 1025)
        self.assertEqual(compress[n], 16368)
        # Adding another name should not increase the size of the compression
        # table, as the pointer would be at offset 16384, which is too big.
        n = dns.name.from_text("toobig.com.")
        n.to_wire(f, compress)
        self.assertEqual(len(compress), 1025)

    def testSplit1(self):
        n = dns.name.from_text("foo.bar.")
        (prefix, suffix) = n.split(2)
        ep = dns.name.from_text("foo", None)
        es = dns.name.from_text("bar.", None)
        self.assertEqual(prefix, ep)
        self.assertEqual(suffix, es)

    def testSplit2(self):
        n = dns.name.from_text("foo.bar.")
        (prefix, suffix) = n.split(1)
        ep = dns.name.from_text("foo.bar", None)
        es = dns.name.from_text(".", None)
        self.assertEqual(prefix, ep)
        self.assertEqual(suffix, es)

    def testSplit3(self):
        n = dns.name.from_text("foo.bar.")
        (prefix, suffix) = n.split(0)
        ep = dns.name.from_text("foo.bar.", None)
        es = dns.name.from_text("", None)
        self.assertEqual(prefix, ep)
        self.assertEqual(suffix, es)

    def testSplit4(self):
        n = dns.name.from_text("foo.bar.")
        (prefix, suffix) = n.split(3)
        ep = dns.name.from_text("", None)
        es = dns.name.from_text("foo.bar.", None)
        self.assertEqual(prefix, ep)
        self.assertEqual(suffix, es)

    def testBadSplit1(self):
        def bad():
            n = dns.name.from_text("foo.bar.")
            n.split(-1)

        self.assertRaises(ValueError, bad)

    def testBadSplit2(self):
        def bad():
            n = dns.name.from_text("foo.bar.")
            n.split(4)

        self.assertRaises(ValueError, bad)

    def testRelativize1(self):
        n = dns.name.from_text("a.foo.bar.", None)
        o = dns.name.from_text("bar.", None)
        e = dns.name.from_text("a.foo", None)
        self.assertEqual(n.relativize(o), e)

    def testRelativize2(self):
        n = dns.name.from_text("a.foo.bar.", None)
        o = n
        e = dns.name.empty
        self.assertEqual(n.relativize(o), e)

    def testRelativize3(self):
        n = dns.name.from_text("a.foo.bar.", None)
        o = dns.name.from_text("blaz.", None)
        e = n
        self.assertEqual(n.relativize(o), e)

    def testRelativize4(self):
        n = dns.name.from_text("a.foo", None)
        o = dns.name.root
        e = n
        self.assertEqual(n.relativize(o), e)

    def testDerelativize1(self):
        n = dns.name.from_text("a.foo", None)
        o = dns.name.from_text("bar.", None)
        e = dns.name.from_text("a.foo.bar.", None)
        self.assertEqual(n.derelativize(o), e)

    def testDerelativize2(self):
        n = dns.name.empty
        o = dns.name.from_text("a.foo.bar.", None)
        e = o
        self.assertEqual(n.derelativize(o), e)

    def testDerelativize3(self):
        n = dns.name.from_text("a.foo.bar.", None)
        o = dns.name.from_text("blaz.", None)
        e = n
        self.assertEqual(n.derelativize(o), e)

    def testChooseRelativity1(self):
        n = dns.name.from_text("a.foo.bar.", None)
        o = dns.name.from_text("bar.", None)
        e = dns.name.from_text("a.foo", None)
        self.assertEqual(n.choose_relativity(o, True), e)

    def testChooseRelativity2(self):
        n = dns.name.from_text("a.foo.bar.", None)
        o = dns.name.from_text("bar.", None)
        e = n
        self.assertEqual(n.choose_relativity(o, False), e)

    def testChooseRelativity3(self):
        n = dns.name.from_text("a.foo", None)
        o = dns.name.from_text("bar.", None)
        e = dns.name.from_text("a.foo.bar.", None)
        self.assertEqual(n.choose_relativity(o, False), e)

    def testChooseRelativity4(self):
        n = dns.name.from_text("a.foo", None)
        o = None
        e = n
        self.assertEqual(n.choose_relativity(o, True), e)

    def testChooseRelativity5(self):
        n = dns.name.from_text("a.foo", None)
        o = None
        e = n
        self.assertEqual(n.choose_relativity(o, False), e)

    def testChooseRelativity6(self):
        n = dns.name.from_text("a.foo.", None)
        o = None
        e = n
        self.assertEqual(n.choose_relativity(o, True), e)

    def testChooseRelativity7(self):
        n = dns.name.from_text("a.foo.", None)
        o = None
        e = n
        self.assertEqual(n.choose_relativity(o, False), e)

    def testFromWire1(self):
        w = b"\x03foo\x00\xc0\x00"
        (n1, cused1) = dns.name.from_wire(w, 0)
        (n2, cused2) = dns.name.from_wire(w, cused1)
        en1 = dns.name.from_text("foo.")
        en2 = en1
        ecused1 = 5
        ecused2 = 2
        self.assertEqual(n1, en1)
        self.assertEqual(cused1, ecused1)
        self.assertEqual(n2, en2)
        self.assertEqual(cused2, ecused2)

    def testFromWire2(self):
        w = b"\x03foo\x00\x01a\xc0\x00\x01b\xc0\x05"
        current = 0
        (n1, cused1) = dns.name.from_wire(w, current)
        current += cused1
        (n2, cused2) = dns.name.from_wire(w, current)
        current += cused2
        (n3, cused3) = dns.name.from_wire(w, current)
        en1 = dns.name.from_text("foo.")
        en2 = dns.name.from_text("a.foo.")
        en3 = dns.name.from_text("b.a.foo.")
        ecused1 = 5
        ecused2 = 4
        ecused3 = 4
        self.assertEqual(n1, en1)
        self.assertEqual(cused1, ecused1)
        self.assertEqual(n2, en2)
        self.assertEqual(cused2, ecused2)
        self.assertEqual(n3, en3)
        self.assertEqual(cused3, ecused3)

    def testBadFromWire1(self):
        def bad():
            w = b"\x03foo\xc0\x04"
            dns.name.from_wire(w, 0)

        self.assertRaises(dns.name.BadPointer, bad)

    def testBadFromWire2(self):
        def bad():
            w = b"\x03foo\xc0\x05"
            dns.name.from_wire(w, 0)

        self.assertRaises(dns.name.BadPointer, bad)

    def testBadFromWire3(self):
        def bad():
            w = b"\xbffoo"
            dns.name.from_wire(w, 0)

        self.assertRaises(dns.name.BadLabelType, bad)

    def testBadFromWire4(self):
        def bad():
            w = b"\x41foo"
            dns.name.from_wire(w, 0)

        self.assertRaises(dns.name.BadLabelType, bad)

    def testParent1(self):
        n = dns.name.from_text("foo.bar.")
        self.assertEqual(n.parent(), dns.name.from_text("bar."))
        self.assertEqual(n.parent().parent(), dns.name.root)

    def testParent2(self):
        n = dns.name.from_text("foo.bar", None)
        self.assertEqual(n.parent(), dns.name.from_text("bar", None))
        self.assertEqual(n.parent().parent(), dns.name.empty)

    def testParent3(self):
        def bad():
            n = dns.name.root
            n.parent()

        self.assertRaises(dns.name.NoParent, bad)

    def testParent4(self):
        def bad():
            n = dns.name.empty
            n.parent()

        self.assertRaises(dns.name.NoParent, bad)

    def testFromUnicode1(self):
        n = dns.name.from_text("foo.bar")
        self.assertEqual(n.labels, (b"foo", b"bar", b""))

    def testFromUnicode2(self):
        n = dns.name.from_text("foo\u1234bar.bar")
        self.assertEqual(n.labels, (b"xn--foobar-r5z", b"bar", b""))

    def testFromUnicodeAlternateDot1(self):
        n = dns.name.from_text("foo\u3002bar")
        self.assertEqual(n.labels, (b"foo", b"bar", b""))

    def testFromUnicodeAlternateDot2(self):
        n = dns.name.from_text("foo\uff0ebar")
        self.assertEqual(n.labels, (b"foo", b"bar", b""))

    def testFromUnicodeAlternateDot3(self):
        n = dns.name.from_text("foo\uff61bar")
        self.assertEqual(n.labels, (b"foo", b"bar", b""))

    def testFromUnicodeRoot(self):
        n = dns.name.from_text(".")
        self.assertEqual(n.labels, (b"",))

    def testFromUnicodeAlternateRoot1(self):
        n = dns.name.from_text("\u3002")
        self.assertEqual(n.labels, (b"",))

    def testFromUnicodeAlternateRoot2(self):
        n = dns.name.from_text("\uff0e")
        self.assertEqual(n.labels, (b"",))

    def testFromUnicodeAlternateRoot3(self):
        n = dns.name.from_text("\uff61")
        self.assertEqual(n.labels, (b"",))

    def testFromUnicodeIDNA2003Explicit(self):
        t = "Königsgäßchen"
        e = dns.name.from_unicode(t, idna_codec=dns.name.IDNA_2003)
        self.assertEqual(str(e), "xn--knigsgsschen-lcb0w.")

    def testFromUnicodeIDNA2003Default(self):
        t = "Königsgäßchen"
        e = dns.name.from_unicode(t)
        self.assertEqual(str(e), "xn--knigsgsschen-lcb0w.")

    @unittest.skipUnless(
        dns.name.have_idna_2008, "Python idna cannot be imported; no IDNA2008"
    )
    def testFromUnicodeIDNA2008(self):
        t = "Königsgäßchen"

        def bad():
            codec = dns.name.IDNA_2008_Strict
            return dns.name.from_unicode(t, idna_codec=codec)

        self.assertRaises(dns.name.IDNAException, bad)
        e1 = dns.name.from_unicode(t, idna_codec=dns.name.IDNA_2008)
        self.assertEqual(str(e1), "xn--knigsgchen-b4a3dun.")
        c2 = dns.name.IDNA_2008_Transitional
        e2 = dns.name.from_unicode(t, idna_codec=c2)
        self.assertEqual(str(e2), "xn--knigsgsschen-lcb0w.")

    @unittest.skipUnless(
        dns.name.have_idna_2008, "Python idna cannot be imported; no IDNA2008"
    )
    def testFromUnicodeIDNA2008Mixed(self):
        # the IDN rules for names are very restrictive, disallowing
        # practical names like '_sip._tcp.Königsgäßchen'.  Dnspython
        # has a "practical" mode which permits labels which are purely
        # ASCII to go straight through, and thus not invalid useful
        # things in the real world.
        t = "_sip._tcp.Königsgäßchen"

        def bad1():
            codec = dns.name.IDNA_2008_Strict
            return dns.name.from_unicode(t, idna_codec=codec)

        def bad2():
            codec = dns.name.IDNA_2008_UTS_46
            return dns.name.from_unicode(t, idna_codec=codec)

        def bad3():
            codec = dns.name.IDNA_2008_Transitional
            return dns.name.from_unicode(t, idna_codec=codec)

        self.assertRaises(dns.name.IDNAException, bad1)
        self.assertRaises(dns.name.IDNAException, bad2)
        self.assertRaises(dns.name.IDNAException, bad3)
        e = dns.name.from_unicode(t, idna_codec=dns.name.IDNA_2008_Practical)
        self.assertEqual(str(e), "_sip._tcp.xn--knigsgchen-b4a3dun.")

    def testFromUnicodeEscapes(self):
        n = dns.name.from_unicode(r"\097.\098.\099.")
        t = n.to_unicode()
        self.assertEqual(t, "a.b.c.")

    def testToUnicode1(self):
        n = dns.name.from_text("foo.bar")
        s = n.to_unicode()
        self.assertEqual(s, "foo.bar.")

    def testToUnicode2(self):
        n = dns.name.from_text("foo\u1234bar.bar")
        s = n.to_unicode()
        self.assertEqual(s, "foo\u1234bar.bar.")

    def testToUnicode3(self):
        n = dns.name.from_text("foo.bar")
        s = n.to_unicode()
        self.assertEqual(s, "foo.bar.")

    @unittest.skipUnless(
        dns.name.have_idna_2008, "Python idna cannot be imported; no IDNA2008"
    )
    def testToUnicode4(self):
        n = dns.name.from_text("ドメイン.テスト", idna_codec=dns.name.IDNA_2008)
        s = n.to_unicode()
        self.assertEqual(str(n), "xn--eckwd4c7c.xn--zckzah.")
        self.assertEqual(s, "ドメイン.テスト.")

    @unittest.skipUnless(
        dns.name.have_idna_2008, "Python idna cannot be imported; no IDNA2008"
    )
    def testToUnicode5(self):
        # Exercise UTS 46 remapping in decode.  This doesn't normally happen
        # as you can see from us having to instantiate the codec as
        # transitional with strict decoding, not one of our usual choices.
        codec = dns.name.IDNA2008Codec(True, True, False, True)
        n = dns.name.from_text("xn--gro-7ka.com")
        self.assertEqual(n.to_unicode(idna_codec=codec), "gross.com.")

    @unittest.skipUnless(
        dns.name.have_idna_2008, "Python idna cannot be imported; no IDNA2008"
    )
    def testToUnicode6(self):
        # Test strict 2008 decoding without UTS 46
        n = dns.name.from_text("xn--gro-7ka.com")
        self.assertEqual(
            n.to_unicode(idna_codec=dns.name.IDNA_2008_Strict), "groß.com."
        )

    def testDefaultDecodeIsJustPunycode(self):
        # groß.com. in IDNA2008 form, pre-encoded.
        n = dns.name.from_text("xn--gro-7ka.com")
        # output using default codec which just decodes the punycode and
        # doesn't test for IDNA2003 or IDNA2008.
        self.assertEqual(n.to_unicode(), "groß.com.")

    def testStrictINDA2003Decode(self):
        # groß.com. in IDNA2008 form, pre-encoded.
        n = dns.name.from_text("xn--gro-7ka.com")

        def bad():
            # This throws in IDNA2003 because it doesn't "round trip".
            n.to_unicode(idna_codec=dns.name.IDNA_2003_Strict)

        self.assertRaises(dns.name.IDNAException, bad)

    def testINDA2008Decode(self):
        # groß.com. in IDNA2008 form, pre-encoded.
        n = dns.name.from_text("xn--gro-7ka.com")
        self.assertEqual(n.to_unicode(idna_codec=dns.name.IDNA_2008), "groß.com.")

    def testToUnicodeOmitFinalDot(self):
        # groß.com. in IDNA2008 form, pre-encoded.
        n = dns.name.from_text("xn--gro-7ka.com")
        self.assertEqual(n.to_unicode(True, dns.name.IDNA_2008), "groß.com")

    def testIDNA2003Misc(self):
        self.assertEqual(dns.name.IDNA_2003.encode(""), b"")
        self.assertRaises(
            dns.name.LabelTooLong, lambda: dns.name.IDNA_2003.encode("x" * 64)
        )

    @unittest.skipUnless(
        dns.name.have_idna_2008, "Python idna cannot be imported; no IDNA2008"
    )
    def testIDNA2008Misc(self):
        self.assertEqual(dns.name.IDNA_2008.encode(""), b"")
        self.assertRaises(
            dns.name.LabelTooLong, lambda: dns.name.IDNA_2008.encode("x" * 64)
        )
        self.assertRaises(
            dns.name.LabelTooLong, lambda: dns.name.IDNA_2008.encode("groß" + "x" * 60)
        )

    def testReverseIPv4(self):
        e = dns.name.from_text("1.0.0.127.in-addr.arpa.")
        n = dns.reversename.from_address("127.0.0.1")
        self.assertEqual(e, n)

    def testReverseIPv6(self):
        e = dns.name.from_text(
            "1.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.ip6.arpa."
        )
        n = dns.reversename.from_address("::1")
        self.assertEqual(e, n)

    def testReverseIPv6MappedIpv4(self):
        e = dns.name.from_text("1.0.0.127.in-addr.arpa.")
        n = dns.reversename.from_address("::ffff:127.0.0.1")
        self.assertEqual(e, n)

    def testBadReverseIPv4(self):
        def bad():
            dns.reversename.from_address("127.0.foo.1")

        self.assertRaises(dns.exception.SyntaxError, bad)

    def testBadReverseIPv6(self):
        def bad():
            dns.reversename.from_address("::1::1")

        self.assertRaises(dns.exception.SyntaxError, bad)

    def testReverseIPv4AlternateOrigin(self):
        e = dns.name.from_text("1.0.0.127.foo.bar.")
        origin = dns.name.from_text("foo.bar")
        n = dns.reversename.from_address("127.0.0.1", v4_origin=origin)
        self.assertEqual(e, n)

    def testReverseIPv6AlternateOrigin(self):
        e = dns.name.from_text(
            "1.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.foo.bar."
        )
        origin = dns.name.from_text("foo.bar")
        n = dns.reversename.from_address("::1", v6_origin=origin)
        self.assertEqual(e, n)

    def testForwardIPv4(self):
        n = dns.name.from_text("1.0.0.127.in-addr.arpa.")
        e = "127.0.0.1"
        text = dns.reversename.to_address(n)
        self.assertEqual(text, e)

    def testForwardIPv6(self):
        n = dns.name.from_text(
            "1.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.ip6.arpa."
        )
        e = "::1"
        text = dns.reversename.to_address(n)
        self.assertEqual(text, e)

    def testForwardIPv4AlternateOrigin(self):
        n = dns.name.from_text("1.0.0.127.foo.bar.")
        e = "127.0.0.1"
        origin = dns.name.from_text("foo.bar")
        text = dns.reversename.to_address(n, v4_origin=origin)
        self.assertEqual(text, e)

    def testForwardIPv6AlternateOrigin(self):
        n = dns.name.from_text(
            "1.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.foo.bar."
        )
        e = "::1"
        origin = dns.name.from_text("foo.bar")
        text = dns.reversename.to_address(n, v6_origin=origin)
        self.assertEqual(text, e)

    def testUnknownReverseOrigin(self):
        n = dns.name.from_text("1.2.3.4.unknown.")
        with self.assertRaises(dns.exception.SyntaxError):
            dns.reversename.to_address(n)

    def testE164ToEnum(self):
        text = "+1 650 555 1212"
        e = dns.name.from_text("2.1.2.1.5.5.5.0.5.6.1.e164.arpa.")
        n = dns.e164.from_e164(text)
        self.assertEqual(n, e)

    def testEnumToE164(self):
        n = dns.name.from_text("2.1.2.1.5.5.5.0.5.6.1.e164.arpa.")
        e = "+16505551212"
        text = dns.e164.to_e164(n)
        self.assertEqual(text, e)

    def testBadEnumToE164(self):
        n = dns.name.from_text("2.1.2.q.5.5.5.0.5.6.1.e164.arpa.")
        self.assertRaises(dns.exception.SyntaxError, lambda: dns.e164.to_e164(n))

    def test_incompatible_relations(self):
        n1 = dns.name.from_text("example")
        n2 = "abc"
        for oper in [operator.lt, operator.le, operator.ge, operator.gt]:
            self.assertRaises(TypeError, lambda: oper(n1, n2))
        self.assertFalse(n1 == n2)
        self.assertTrue(n1 != n2)

    def testFromUnicodeSimpleEscape(self):
        n = dns.name.from_unicode(r"a.\b")
        e = dns.name.from_unicode(r"a.b")
        self.assertEqual(n, e)

    def testFromUnicodeBadEscape(self):
        def bad1():
            n = dns.name.from_unicode(r"a.b\0q1.c.")

        self.assertRaises(dns.name.BadEscape, bad1)

        def bad2():
            n = dns.name.from_unicode(r"a.b\0")

        self.assertRaises(dns.name.BadEscape, bad2)

    def testFromUnicodeNotString(self):
        def bad():
            dns.name.from_unicode(b"123")  # type: ignore

        self.assertRaises(ValueError, bad)

    def testFromUnicodeBadOrigin(self):
        def bad():
            dns.name.from_unicode("example", 123)  # type: ignore

        self.assertRaises(ValueError, bad)

    def testFromUnicodeEmptyLabel(self):
        def bad():
            dns.name.from_unicode("a..b.example")

        self.assertRaises(dns.name.EmptyLabel, bad)

    def testFromUnicodeEmptyName(self):
        self.assertEqual(dns.name.from_unicode("@", None), dns.name.empty)

    def testFromTextNotString(self):
        def bad():
            dns.name.from_text(123)  # type: ignore

        self.assertRaises(ValueError, bad)

    def testFromTextBadOrigin(self):
        def bad():
            dns.name.from_text("example", 123)  # type: ignore

        self.assertRaises(ValueError, bad)

    def testFromWireNotBytes(self):
        def bad():
            dns.name.from_wire(123, 0)  # type: ignore

        self.assertRaises(ValueError, bad)

    def testBadPunycode(self):
        c = dns.name.IDNACodec()
        with self.assertRaises(dns.name.IDNAException):
            c.decode(b"xn--0000h")

    def testRootLabel2003StrictDecode(self):
        c = dns.name.IDNA_2003_Strict
        self.assertEqual(c.decode(b""), "")

    @unittest.skipUnless(
        dns.name.have_idna_2008, "Python idna cannot be imported; no IDNA2008"
    )
    def testRootLabel2008StrictDecode(self):
        c = dns.name.IDNA_2008_Strict
        self.assertEqual(c.decode(b""), "")

    @unittest.skipUnless(
        dns.name.have_idna_2008, "Python idna cannot be imported; no IDNA2008"
    )
    def testCodecNotFoundRaises(self):
        dns.name.have_idna_2008 = False
        with self.assertRaises(dns.name.NoIDNA2008):
            c = dns.name.IDNA2008Codec()
            c.encode("Königsgäßchen")
        with self.assertRaises(dns.name.NoIDNA2008):
            c = dns.name.IDNA2008Codec(strict_decode=True)
            c.decode(b"xn--eckwd4c7c.xn--zckzah.")
        dns.name.have_idna_2008 = True

    @unittest.skipUnless(
        dns.name.have_idna_2008, "Python idna cannot be imported; no IDNA2008"
    )
    def testBadPunycodeStrict2008(self):
        c = dns.name.IDNA2008Codec(strict_decode=True)
        with self.assertRaises(dns.name.IDNAException):
            c.decode(b"xn--0000h")

    def testRelativizeSubtractionSyntax(self):
        n = dns.name.from_text("foo.example.")
        o = dns.name.from_text("example.")
        e = dns.name.from_text("foo", None)
        self.assertEqual(n - o, e)

    def testCopy(self):
        n1 = dns.name.from_text("foo.example.")
        n2 = copy.copy(n1)
        self.assertTrue(n1 is not n2)
        # the Name constructor always copies labels, so there is no
        # difference between copy and deepcopy
        self.assertTrue(n1.labels is not n2.labels)
        self.assertEqual(len(n1.labels), len(n2.labels))
        for i, l in enumerate(n1.labels):
            self.assertTrue(l is n2[i])

    def testDeepCopy(self):
        n1 = dns.name.from_text("foo.example.")
        n2 = copy.deepcopy(n1)
        self.assertTrue(n1 is not n2)
        self.assertTrue(n1.labels is not n2.labels)
        self.assertEqual(len(n1.labels), len(n2.labels))
        for i, l in enumerate(n1.labels):
            self.assertTrue(l is n2[i])

    def testNoAttributeDeletion(self):
        n = dns.name.from_text("foo.example.")
        with self.assertRaises(TypeError):
            del n.labels

    def testUnicodeEscapify(self):
        n = dns.name.from_unicode("Königsgäßchen;\ttext")
        self.assertEqual(n.to_unicode(), "königsgässchen\\;\\009text.")

    def test_pickle(self):
        n1 = dns.name.from_text("foo.example")
        p = pickle.dumps(n1)
        n2 = pickle.loads(p)
        self.assertEqual(n1, n2)

    def test_pad_to_max_name(self):
        # Test edge cases in our padding helper.
        tests = [
            ("o{61}.o{63}.o{63}.o{63}.", "o{61}.o{63}.o{63}.o{63}."),
            ("o{60}.o{63}.o{63}.o{63}.", "o{60}.o{63}.o{63}.o{63}."),
            ("o{59}.o{63}.o{63}.o{63}.", "+.o{59}.o{63}.o{63}.o{63}."),
            ("o{63}.o{63}.o{63}.", "+{61}.o{63}.o{63}.o{63}."),
            ("o{63}.o{63}.", "+{61}.+{63}.o{63}.o{63}."),
        ]
        for name_text, expected_text in tests:
            name = dns.name.from_text(expand(name_text))
            expected = dns.name.from_text(expand(expected_text))
            self.assertEqual(dns.name._pad_to_max_name(name), expected)

    def test_predecessors_and_successors(self):
        # Test RFC 4471 predecessor and successor methods.

        # Here we're actually testing the test suite, but expand() is complicated enough
        # to deserve a little testing.
        self.assertEqual(expand("f+{3}-o.example."), r"f\255\255\255\000o.example.")

        # Ok, now test successors!
        origin = dns.name.from_text("example.com.")
        tests = [
            # Examples from the RFC.
            ("foo", True, "\\000.foo"),
            ("foo", False, "foo\\000"),
            # The syntax here is almost the RFC's "alternate syntax"  except that to
            # make expand simpler, i.e. not have to understand that \000 was one octet,
            # I made the convention that "-" means 0 and "+" means 255.  We use raw
            # string constants where needed to avoid escaping backslash.
            (
                "fo{47}.o{63}.o{63}.o{63}",
                True,
                "fo{47}-.o{63}.o{63}.o{63}",
            ),
            (
                "fo{48}.o{63}.o{63}.o{63}",
                True,
                "fo{47}p.o{63}.o{63}.o{63}",
            ),
            ("+{49}.o{63}.o{63}.o{63}", True, "o{62}p.o{63}.o{63}"),
            (
                "fo{40}+{8}.o{63}.o{63}.o{63}",
                True,
                "fo{39}p.o{63}.o{63}.o{63}",
            ),
            (
                r"fo{47}\@.o{63}.o{63}.o{63}",
                True,
                r"fo{47}\[.o{63}.o{63}.o{63}",
            ),
            ("+{49}.+{63}.+{63}.+{63}", True, ""),
            # Some more tests not in the RFC
            ("+{49}.+{63}.o{63}.o{63}", True, "o{62}p.o{63}"),
            (
                "+{48}.o{63}.o{63}.o{63}",
                True,
                "+{48}-.o{63}.o{63}.o{63}",
            ),
        ]
        for test_origin in [origin, None]:
            for name_text, prefix_ok, expected_successor_text in tests:
                name = dns.name.from_text(expand(name_text), test_origin)
                expected_successor = dns.name.from_text(
                    expand(expected_successor_text), test_origin
                )
                successor = name.successor(origin, prefix_ok)
                self.assertEqual(successor, expected_successor)
                self.assertTrue(
                    successor > name
                    or successor == origin
                    or successor == dns.name.empty
                )
                # Now test the predecessor
                predecessor = successor.predecessor(origin, prefix_ok)
                self.assertEqual(predecessor, name)

        # Finally, test that a maximal length origin is its own predecessor and
        # successor.
        origin = dns.name.from_text(expand("+{49}.+{63}.o{63}.o{63}.example.com."))
        assert origin.successor(origin, True) == origin
        assert origin.predecessor(origin, True) == origin

    def test_predecessor_and_successor_errors(self):
        name = dns.name.from_text("name", None)
        origin = dns.name.from_text("origin", None)  # note Relative!
        with self.assertRaises(dns.name.NeedAbsoluteNameOrOrigin):
            name.successor(origin, True)
        with self.assertRaises(dns.name.NeedAbsoluteNameOrOrigin):
            name.predecessor(origin, True)
        name = dns.name.from_text("name.")  # Note absolute
        origin = dns.name.from_text("origin.")  # 'name' is not a subdomain of 'origin'
        with self.assertRaises(dns.name.NeedSubdomainOfOrigin):
            name.successor(origin, True)
        with self.assertRaises(dns.name.NeedSubdomainOfOrigin):
            name.predecessor(origin, True)


if __name__ == "__main__":
    unittest.main()



# Copyright (C) Dnspython Contributors, see LICENSE for text of ISC license

try:
    import asyncio
    import socket
    import struct
    import threading

    import aioquic.asyncio
    import aioquic.asyncio.server
    import aioquic.quic.configuration
    import aioquic.quic.events

    import dns.asyncquery
    import dns.message
    import dns.rcode
    from tests.util import here

    have_quic = True

    class Request:
        def __init__(self, message, wire):
            self.message = message
            self.wire = wire

        @property
        def question(self):
            return self.message.question[0]

        @property
        def qname(self):
            return self.question.name

        @property
        def qclass(self):
            return self.question.rdclass

        @property
        def qtype(self):
            return self.question.rdtype

    class NanoQuic(aioquic.asyncio.QuicConnectionProtocol):
        def quic_event_received(self, event):
            # This is a bit hackish and not fully general, but this is a test server!
            if isinstance(event, aioquic.quic.events.StreamDataReceived):
                data = bytes(event.data)
                (wire_len,) = struct.unpack("!H", data[:2])
                wire = self.handle_wire(data[2 : 2 + wire_len])
                if wire is not None:
                    self._quic.send_stream_data(event.stream_id, wire, end_stream=True)

        def handle(self, request):
            r = dns.message.make_response(request.message)
            r.set_rcode(dns.rcode.REFUSED)
            return r

        def handle_wire(self, wire):
            response = None
            try:
                q = dns.message.from_wire(wire)
            except dns.message.ShortHeader:
                return
            except Exception as e:
                try:
                    q = dns.message.from_wire(wire, question_only=True)
                    response = dns.message.make_response(q)
                    response.set_rcode(dns.rcode.FORMERR)
                except Exception:
                    return
            if response is None:
                try:
                    request = Request(q, wire)
                    response = self.handle(request)
                except Exception:
                    response = dns.message.make_response(q)
                    response.set_rcode(dns.rcode.SERVFAIL)
            wire = response.to_wire()
            return struct.pack("!H", len(wire)) + wire

    class Server(threading.Thread):
        def __init__(self, address="127.0.0.1"):
            super().__init__()
            self.address = address
            self.transport = None
            self.protocol = None
            self.left = None
            self.right = None
            self.ready = threading.Event()

        def __enter__(self):
            self.left, self.right = socket.socketpair()
            self.start()
            self.ready.wait(4)
            return self

        def __exit__(self, ex_ty, ex_va, ex_tr):
            if self.protocol is not None:
                self.protocol.close()
            if self.transport is not None:
                self.transport.close()
            if self.left:
                self.left.close()
            if self.is_alive():
                self.join()
            if self.right:
                self.right.close()

        async def arun(self):
            reader, _ = await asyncio.open_connection(sock=self.right)
            conf = aioquic.quic.configuration.QuicConfiguration(
                alpn_protocols=["doq"],
                is_client=False,
            )
            conf.load_cert_chain(here("tls/public.crt"), here("tls/private.pem"))
            loop = asyncio.get_event_loop()
            (self.transport, self.protocol) = await loop.create_datagram_endpoint(
                lambda: aioquic.asyncio.server.QuicServer(
                    configuration=conf, create_protocol=NanoQuic
                ),
                local_addr=(self.address, 0),
            )
            info = self.transport.get_extra_info("sockname")
            self.port = info[1]
            self.ready.set()
            try:
                await reader.read(1)
            except Exception:
                pass

        def run(self):
            asyncio.run(self.arun())

except ImportError:
    have_quic = False

    class NanoQuic:
        pass



# Copyright (C) Dnspython Contributors, see LICENSE for text of ISC license

import asyncio
import sys

import pytest

import dns.asyncbackend
import dns.asyncquery
import dns.message
import dns.query
import dns.tsigkeyring
import dns.versioned
import dns.xfr

# Some tests use a "nano nameserver" for testing.  It requires trio
# and threading, so try to import it and if it doesn't work, skip
# those tests.
try:
    from .nanonameserver import Server

    _nanonameserver_available = True
except ImportError:
    _nanonameserver_available = False

    class Server(object):
        pass


axfr = """id 1
opcode QUERY
rcode NOERROR
flags AA
;QUESTION
example. IN AXFR
;ANSWER
@ 3600 IN SOA foo bar 1 2 3 4 5
@ 3600 IN NS ns1
@ 3600 IN NS ns2
bar.foo 300 IN MX 0 blaz.foo
ns1 3600 IN A 10.0.0.1
ns2 3600 IN A 10.0.0.2
@ 3600 IN SOA foo bar 1 2 3 4 5
"""

axfr1 = """id 1
opcode QUERY
rcode NOERROR
flags AA
;QUESTION
example. IN AXFR
;ANSWER
@ 3600 IN SOA foo bar 1 2 3 4 5
@ 3600 IN NS ns1
@ 3600 IN NS ns2
"""
axfr2 = """id 1
opcode QUERY
rcode NOERROR
flags AA
;ANSWER
bar.foo 300 IN MX 0 blaz.foo
ns1 3600 IN A 10.0.0.1
ns2 3600 IN A 10.0.0.2
@ 3600 IN SOA foo bar 1 2 3 4 5
"""

base = """@ 3600 IN SOA foo bar 1 2 3 4 5
@ 3600 IN NS ns1
@ 3600 IN NS ns2
bar.foo 300 IN MX 0 blaz.foo
ns1 3600 IN A 10.0.0.1
ns2 3600 IN A 10.0.0.2
"""

axfr_unexpected_origin = """id 1
opcode QUERY
rcode NOERROR
flags AA
;QUESTION
example. IN AXFR
;ANSWER
@ 3600 IN SOA foo bar 1 2 3 4 5
@ 3600 IN SOA foo bar 1 2 3 4 7
"""

ixfr = """id 1
opcode QUERY
rcode NOERROR
flags AA
;QUESTION
example. IN IXFR
;ANSWER
@ 3600 IN SOA foo bar 4 2 3 4 5
@ 3600 IN SOA foo bar 1 2 3 4 5
bar.foo 300 IN MX 0 blaz.foo
ns2 3600 IN A 10.0.0.2
@ 3600 IN SOA foo bar 2 2 3 4 5
ns2 3600 IN A 10.0.0.4
@ 3600 IN SOA foo bar 2 2 3 4 5
@ 3600 IN SOA foo bar 3 2 3 4 5
ns3 3600 IN A 10.0.0.3
@ 3600 IN SOA foo bar 3 2 3 4 5
@ 3600 IN NS ns2
@ 3600 IN SOA foo bar 4 2 3 4 5
@ 3600 IN SOA foo bar 4 2 3 4 5
"""

compressed_ixfr = """id 1
opcode QUERY
rcode NOERROR
flags AA
;QUESTION
example. IN IXFR
;ANSWER
@ 3600 IN SOA foo bar 4 2 3 4 5
@ 3600 IN SOA foo bar 1 2 3 4 5
bar.foo 300 IN MX 0 blaz.foo
ns2 3600 IN A 10.0.0.2
@ 3600 IN NS ns2
@ 3600 IN SOA foo bar 4 2 3 4 5
ns2 3600 IN A 10.0.0.4
ns3 3600 IN A 10.0.0.3
@ 3600 IN SOA foo bar 4 2 3 4 5
"""

ixfr_expected = """@ 3600 IN SOA foo bar 4 2 3 4 5
@ 3600 IN NS ns1
ns1 3600 IN A 10.0.0.1
ns2 3600 IN A 10.0.0.4
ns3 3600 IN A 10.0.0.3
"""

ixfr_first_message = """id 1
opcode QUERY
rcode NOERROR
flags AA
;QUESTION
example. IN IXFR
;ANSWER
@ 3600 IN SOA foo bar 4 2 3 4 5
"""

ixfr_header = """id 1
opcode QUERY
rcode NOERROR
flags AA
;ANSWER
"""

ixfr_body = [
    "@ 3600 IN SOA foo bar 1 2 3 4 5",
    "bar.foo 300 IN MX 0 blaz.foo",
    "ns2 3600 IN A 10.0.0.2",
    "@ 3600 IN SOA foo bar 2 2 3 4 5",
    "ns2 3600 IN A 10.0.0.4",
    "@ 3600 IN SOA foo bar 2 2 3 4 5",
    "@ 3600 IN SOA foo bar 3 2 3 4 5",
    "ns3 3600 IN A 10.0.0.3",
    "@ 3600 IN SOA foo bar 3 2 3 4 5",
    "@ 3600 IN NS ns2",
    "@ 3600 IN SOA foo bar 4 2 3 4 5",
    "@ 3600 IN SOA foo bar 4 2 3 4 5",
]

ixfrs = [ixfr_first_message]
ixfrs.extend([ixfr_header + l for l in ixfr_body])

good_empty_ixfr = """id 1
opcode QUERY
rcode NOERROR
flags AA
;QUESTION
example. IN IXFR
;ANSWER
@ 3600 IN SOA foo bar 1 2 3 4 5
"""

retry_tcp_ixfr = """id 1
opcode QUERY
rcode NOERROR
flags AA
;QUESTION
example. IN IXFR
;ANSWER
@ 3600 IN SOA foo bar 5 2 3 4 5
"""

bad_empty_ixfr = """id 1
opcode QUERY
rcode NOERROR
flags AA
;QUESTION
example. IN IXFR
;ANSWER
@ 3600 IN SOA foo bar 4 2 3 4 5
@ 3600 IN SOA foo bar 4 2 3 4 5
"""

unexpected_end_ixfr = """id 1
opcode QUERY
rcode NOERROR
flags AA
;QUESTION
example. IN IXFR
;ANSWER
@ 3600 IN SOA foo bar 4 2 3 4 5
@ 3600 IN SOA foo bar 1 2 3 4 5
bar.foo 300 IN MX 0 blaz.foo
ns2 3600 IN A 10.0.0.2
@ 3600 IN NS ns2
@ 3600 IN SOA foo bar 3 2 3 4 5
ns2 3600 IN A 10.0.0.4
ns3 3600 IN A 10.0.0.3
@ 3600 IN SOA foo bar 4 2 3 4 5
"""

unexpected_end_ixfr_2 = """id 1
opcode QUERY
rcode NOERROR
flags AA
;QUESTION
example. IN IXFR
;ANSWER
@ 3600 IN SOA foo bar 4 2 3 4 5
@ 3600 IN SOA foo bar 1 2 3 4 5
bar.foo 300 IN MX 0 blaz.foo
ns2 3600 IN A 10.0.0.2
@ 3600 IN NS ns2
"""

bad_serial_ixfr = """id 1
opcode QUERY
rcode NOERROR
flags AA
;QUESTION
example. IN IXFR
;ANSWER
@ 3600 IN SOA foo bar 4 2 3 4 5
@ 3600 IN SOA foo bar 2 2 3 4 5
bar.foo 300 IN MX 0 blaz.foo
ns2 3600 IN A 10.0.0.2
@ 3600 IN NS ns2
@ 3600 IN SOA foo bar 4 2 3 4 5
ns2 3600 IN A 10.0.0.4
ns3 3600 IN A 10.0.0.3
@ 3600 IN SOA foo bar 4 2 3 4 5
"""

ixfr_axfr = """id 1
opcode QUERY
rcode NOERROR
flags AA
;QUESTION
example. IN IXFR
;ANSWER
@ 3600 IN SOA foo bar 1 2 3 4 5
@ 3600 IN NS ns1
@ 3600 IN NS ns2
bar.foo 300 IN MX 0 blaz.foo
ns1 3600 IN A 10.0.0.1
ns2 3600 IN A 10.0.0.2
@ 3600 IN SOA foo bar 1 2 3 4 5
"""


def test_basic_axfr():
    z = dns.versioned.Zone("example.")
    m = dns.message.from_text(axfr, origin=z.origin, one_rr_per_rrset=True)
    with dns.xfr.Inbound(z, dns.rdatatype.AXFR) as xfr:
        done = xfr.process_message(m)
        assert done
    ez = dns.zone.from_text(base, "example.")
    assert z == ez


def test_basic_axfr_unversioned():
    z = dns.zone.Zone("example.")
    m = dns.message.from_text(axfr, origin=z.origin, one_rr_per_rrset=True)
    with dns.xfr.Inbound(z, dns.rdatatype.AXFR) as xfr:
        done = xfr.process_message(m)
        assert done
    ez = dns.zone.from_text(base, "example.")
    assert z == ez


def test_basic_axfr_two_parts():
    z = dns.versioned.Zone("example.")
    m1 = dns.message.from_text(axfr1, origin=z.origin, one_rr_per_rrset=True)
    m2 = dns.message.from_text(axfr2, origin=z.origin, one_rr_per_rrset=True)
    with dns.xfr.Inbound(z, dns.rdatatype.AXFR) as xfr:
        done = xfr.process_message(m1)
        assert not done
        done = xfr.process_message(m2)
        assert done
    ez = dns.zone.from_text(base, "example.")
    assert z == ez


def test_axfr_unexpected_origin():
    z = dns.versioned.Zone("example.")
    m = dns.message.from_text(
        axfr_unexpected_origin, origin=z.origin, one_rr_per_rrset=True
    )
    with dns.xfr.Inbound(z, dns.rdatatype.AXFR) as xfr:
        with pytest.raises(dns.exception.FormError):
            xfr.process_message(m)


def test_basic_ixfr():
    z = dns.zone.from_text(base, "example.", zone_factory=dns.versioned.Zone)
    m = dns.message.from_text(ixfr, origin=z.origin, one_rr_per_rrset=True)
    with dns.xfr.Inbound(z, dns.rdatatype.IXFR, serial=1) as xfr:
        done = xfr.process_message(m)
        assert done
    ez = dns.zone.from_text(ixfr_expected, "example.")
    assert z == ez


def test_basic_ixfr_unversioned():
    z = dns.zone.from_text(base, "example.")
    m = dns.message.from_text(ixfr, origin=z.origin, one_rr_per_rrset=True)
    with dns.xfr.Inbound(z, dns.rdatatype.IXFR, serial=1) as xfr:
        done = xfr.process_message(m)
        assert done
    ez = dns.zone.from_text(ixfr_expected, "example.")
    assert z == ez


def test_compressed_ixfr():
    z = dns.zone.from_text(base, "example.", zone_factory=dns.versioned.Zone)
    m = dns.message.from_text(compressed_ixfr, origin=z.origin, one_rr_per_rrset=True)
    with dns.xfr.Inbound(z, dns.rdatatype.IXFR, serial=1) as xfr:
        done = xfr.process_message(m)
        assert done
    ez = dns.zone.from_text(ixfr_expected, "example.")
    assert z == ez


def test_basic_ixfr_many_parts():
    z = dns.zone.from_text(base, "example.", zone_factory=dns.versioned.Zone)
    with dns.xfr.Inbound(z, dns.rdatatype.IXFR, serial=1) as xfr:
        done = False
        for text in ixfrs:
            assert not done
            m = dns.message.from_text(text, origin=z.origin, one_rr_per_rrset=True)
            done = xfr.process_message(m)
        assert done
    ez = dns.zone.from_text(ixfr_expected, "example.")
    assert z == ez


def test_good_empty_ixfr():
    z = dns.zone.from_text(ixfr_expected, "example.", zone_factory=dns.versioned.Zone)
    m = dns.message.from_text(good_empty_ixfr, origin=z.origin, one_rr_per_rrset=True)
    with dns.xfr.Inbound(z, dns.rdatatype.IXFR, serial=1) as xfr:
        done = xfr.process_message(m)
        assert done
    ez = dns.zone.from_text(ixfr_expected, "example.")
    assert z == ez


def test_retry_tcp_ixfr():
    z = dns.zone.from_text(ixfr_expected, "example.", zone_factory=dns.versioned.Zone)
    m = dns.message.from_text(retry_tcp_ixfr, origin=z.origin, one_rr_per_rrset=True)
    with dns.xfr.Inbound(z, dns.rdatatype.IXFR, serial=1, is_udp=True) as xfr:
        with pytest.raises(dns.xfr.UseTCP):
            xfr.process_message(m)


def test_bad_empty_ixfr():
    z = dns.zone.from_text(ixfr_expected, "example.", zone_factory=dns.versioned.Zone)
    m = dns.message.from_text(bad_empty_ixfr, origin=z.origin, one_rr_per_rrset=True)
    with dns.xfr.Inbound(z, dns.rdatatype.IXFR, serial=3) as xfr:
        with pytest.raises(dns.exception.FormError):
            xfr.process_message(m)


def test_serial_went_backwards_ixfr():
    z = dns.zone.from_text(ixfr_expected, "example.", zone_factory=dns.versioned.Zone)
    m = dns.message.from_text(bad_empty_ixfr, origin=z.origin, one_rr_per_rrset=True)
    with dns.xfr.Inbound(z, dns.rdatatype.IXFR, serial=5) as xfr:
        with pytest.raises(dns.xfr.SerialWentBackwards):
            xfr.process_message(m)


def test_ixfr_is_axfr():
    z = dns.zone.from_text(base, "example.", zone_factory=dns.versioned.Zone)
    m = dns.message.from_text(ixfr_axfr, origin=z.origin, one_rr_per_rrset=True)
    with dns.xfr.Inbound(z, dns.rdatatype.IXFR, serial=0xFFFFFFFF) as xfr:
        done = xfr.process_message(m)
        assert done
    ez = dns.zone.from_text(base, "example.")
    assert z == ez


def test_ixfr_requires_serial():
    z = dns.zone.from_text(base, "example.", zone_factory=dns.versioned.Zone)
    with pytest.raises(ValueError):
        dns.xfr.Inbound(z, dns.rdatatype.IXFR)


def test_ixfr_unexpected_end_bad_diff_sequence():
    # This is where we get the end serial, but haven't seen all of
    # the expected diffs
    z = dns.zone.from_text(base, "example.", zone_factory=dns.versioned.Zone)
    m = dns.message.from_text(
        unexpected_end_ixfr, origin=z.origin, one_rr_per_rrset=True
    )
    with dns.xfr.Inbound(z, dns.rdatatype.IXFR, serial=1) as xfr:
        with pytest.raises(dns.exception.FormError):
            xfr.process_message(m)


def test_udp_ixfr_unexpected_end_just_stops():
    # This is where everything looks good, but the IXFR just stops
    # in the middle.
    z = dns.zone.from_text(base, "example.", zone_factory=dns.versioned.Zone)
    m = dns.message.from_text(
        unexpected_end_ixfr_2, origin=z.origin, one_rr_per_rrset=True
    )
    with dns.xfr.Inbound(z, dns.rdatatype.IXFR, serial=1, is_udp=True) as xfr:
        with pytest.raises(dns.exception.FormError):
            xfr.process_message(m)


def test_ixfr_bad_serial():
    z = dns.zone.from_text(base, "example.", zone_factory=dns.versioned.Zone)
    m = dns.message.from_text(bad_serial_ixfr, origin=z.origin, one_rr_per_rrset=True)
    with dns.xfr.Inbound(z, dns.rdatatype.IXFR, serial=1) as xfr:
        with pytest.raises(dns.exception.FormError):
            xfr.process_message(m)


def test_no_udp_with_axfr():
    z = dns.versioned.Zone("example.")
    with pytest.raises(ValueError):
        with dns.xfr.Inbound(z, dns.rdatatype.AXFR, is_udp=True) as xfr:
            pass


refused = """id 1
opcode QUERY
rcode REFUSED
flags AA
;QUESTION
example. IN AXFR
"""

bad_qname = """id 1
opcode QUERY
rcode NOERROR
flags AA
;QUESTION
not-example. IN IXFR
"""

bad_qtype = """id 1
opcode QUERY
rcode NOERROR
flags AA
;QUESTION
example. IN AXFR
"""

soa_not_first = """id 1
opcode QUERY
rcode NOERROR
flags AA
;QUESTION
example. IN IXFR
;ANSWER
bar.foo 300 IN MX 0 blaz.foo
"""

soa_not_first_2 = """id 1
opcode QUERY
rcode NOERROR
flags AA
;QUESTION
example. IN IXFR
;ANSWER
@ 300 IN MX 0 blaz.foo
"""

no_answer = """id 1
opcode QUERY
rcode NOERROR
flags AA
;QUESTION
example. IN IXFR
;ADDITIONAL
bar.foo 300 IN MX 0 blaz.foo
"""

axfr_answers_after_final_soa = """id 1
opcode QUERY
rcode NOERROR
flags AA
;QUESTION
example. IN AXFR
;ANSWER
@ 3600 IN SOA foo bar 1 2 3 4 5
@ 3600 IN NS ns1
@ 3600 IN NS ns2
bar.foo 300 IN MX 0 blaz.foo
ns1 3600 IN A 10.0.0.1
ns2 3600 IN A 10.0.0.2
@ 3600 IN SOA foo bar 1 2 3 4 5
ns3 3600 IN A 10.0.0.3
"""


def test_refused():
    z = dns.zone.from_text(base, "example.", zone_factory=dns.versioned.Zone)
    m = dns.message.from_text(refused, origin=z.origin, one_rr_per_rrset=True)
    with dns.xfr.Inbound(z, dns.rdatatype.IXFR, serial=1) as xfr:
        with pytest.raises(dns.xfr.TransferError):
            xfr.process_message(m)


def test_bad_qname():
    z = dns.zone.from_text(base, "example.", zone_factory=dns.versioned.Zone)
    m = dns.message.from_text(bad_qname, origin=z.origin, one_rr_per_rrset=True)
    with dns.xfr.Inbound(z, dns.rdatatype.IXFR, serial=1) as xfr:
        with pytest.raises(dns.exception.FormError):
            xfr.process_message(m)


def test_bad_qtype():
    z = dns.zone.from_text(base, "example.", zone_factory=dns.versioned.Zone)
    m = dns.message.from_text(bad_qtype, origin=z.origin, one_rr_per_rrset=True)
    with dns.xfr.Inbound(z, dns.rdatatype.IXFR, serial=1) as xfr:
        with pytest.raises(dns.exception.FormError):
            xfr.process_message(m)


def test_soa_not_first():
    z = dns.zone.from_text(base, "example.", zone_factory=dns.versioned.Zone)
    m = dns.message.from_text(soa_not_first, origin=z.origin, one_rr_per_rrset=True)
    with dns.xfr.Inbound(z, dns.rdatatype.IXFR, serial=1) as xfr:
        with pytest.raises(dns.exception.FormError):
            xfr.process_message(m)
    m = dns.message.from_text(soa_not_first_2, origin=z.origin, one_rr_per_rrset=True)
    with dns.xfr.Inbound(z, dns.rdatatype.IXFR, serial=1) as xfr:
        with pytest.raises(dns.exception.FormError):
            xfr.process_message(m)


def test_no_answer():
    z = dns.zone.from_text(base, "example.", zone_factory=dns.versioned.Zone)
    m = dns.message.from_text(no_answer, origin=z.origin, one_rr_per_rrset=True)
    with dns.xfr.Inbound(z, dns.rdatatype.IXFR, serial=1) as xfr:
        with pytest.raises(dns.exception.FormError):
            xfr.process_message(m)


def test_axfr_answers_after_final_soa():
    z = dns.versioned.Zone("example.")
    m = dns.message.from_text(
        axfr_answers_after_final_soa, origin=z.origin, one_rr_per_rrset=True
    )
    with dns.xfr.Inbound(z, dns.rdatatype.AXFR) as xfr:
        with pytest.raises(dns.exception.FormError):
            xfr.process_message(m)


keyring = dns.tsigkeyring.from_text({"keyname.": "NjHwPsMKjdN++dOfE5iAiQ=="})

keyname = dns.name.from_text("keyname")


def test_make_query_basic():
    z = dns.versioned.Zone("example.")
    (q, s) = dns.xfr.make_query(z)
    assert q.question[0].rdtype == dns.rdatatype.AXFR
    assert s is None
    (q, s) = dns.xfr.make_query(z, serial=None)
    assert q.question[0].rdtype == dns.rdatatype.AXFR
    assert s is None
    (q, s) = dns.xfr.make_query(z, serial=10)
    assert q.question[0].rdtype == dns.rdatatype.IXFR
    assert q.authority[0].rdtype == dns.rdatatype.SOA
    assert q.authority[0][0].serial == 10
    assert s == 10
    with z.writer() as txn:
        txn.add("@", 300, dns.rdata.from_text("in", "soa", ". . 1 2 3 4 5"))
    (q, s) = dns.xfr.make_query(z)
    assert q.question[0].rdtype == dns.rdatatype.IXFR
    assert q.authority[0].rdtype == dns.rdatatype.SOA
    assert q.authority[0][0].serial == 1
    assert s == 1
    (q, s) = dns.xfr.make_query(z, keyring=keyring, keyname=keyname)
    assert q.question[0].rdtype == dns.rdatatype.IXFR
    assert q.authority[0].rdtype == dns.rdatatype.SOA
    assert q.authority[0][0].serial == 1
    assert s == 1
    assert q.keyname == keyname


def test_make_query_bad_serial():
    z = dns.versioned.Zone("example.")
    with pytest.raises(ValueError):
        dns.xfr.make_query(z, serial="hi")
    with pytest.raises(ValueError):
        dns.xfr.make_query(z, serial=-1)
    with pytest.raises(ValueError):
        dns.xfr.make_query(z, serial=4294967296)


def test_extract_serial_from_query():
    z = dns.versioned.Zone("example.")
    (q, s) = dns.xfr.make_query(z)
    xs = dns.xfr.extract_serial_from_query(q)
    assert s is None
    assert s == xs
    (q, s) = dns.xfr.make_query(z, serial=10)
    xs = dns.xfr.extract_serial_from_query(q)
    assert s == 10
    assert s == xs
    q = dns.message.make_query("example", "a")
    with pytest.raises(ValueError):
        dns.xfr.extract_serial_from_query(q)


class XFRNanoNameserver(Server):
    def __init__(self):
        super().__init__(origin=dns.name.from_text("example"))

    def handle(self, request):
        try:
            if request.message.question[0].rdtype == dns.rdatatype.IXFR:
                text = ixfr
            else:
                text = axfr
            r = dns.message.from_text(text, one_rr_per_rrset=True, origin=self.origin)
            r.id = request.message.id
            return r
        except Exception:
            pass


@pytest.mark.skipif(not _nanonameserver_available, reason="requires nanonameserver")
def test_sync_inbound_xfr():
    with XFRNanoNameserver() as ns:
        zone = dns.versioned.Zone("example")
        dns.query.inbound_xfr(
            ns.tcp_address[0],
            zone,
            port=ns.tcp_address[1],
            udp_mode=dns.query.UDPMode.TRY_FIRST,
        )
        dns.query.inbound_xfr(
            ns.tcp_address[0],
            zone,
            port=ns.tcp_address[1],
            udp_mode=dns.query.UDPMode.TRY_FIRST,
        )
        expected = dns.zone.from_text(ixfr_expected, "example")
        assert zone == expected


async def async_inbound_xfr():
    with XFRNanoNameserver() as ns:
        zone = dns.versioned.Zone("example")
        await dns.asyncquery.inbound_xfr(
            ns.tcp_address[0],
            zone,
            port=ns.tcp_address[1],
            udp_mode=dns.query.UDPMode.TRY_FIRST,
        )
        await dns.asyncquery.inbound_xfr(
            ns.tcp_address[0],
            zone,
            port=ns.tcp_address[1],
            udp_mode=dns.query.UDPMode.TRY_FIRST,
        )
        expected = dns.zone.from_text(ixfr_expected, "example")
        assert zone == expected


@pytest.mark.skipif(not _nanonameserver_available, reason="requires nanonameserver")
def test_asyncio_inbound_xfr():
    dns.asyncbackend.set_default_backend("asyncio")

    async def run():
        await async_inbound_xfr()

    asyncio.run(run())


#
# We don't need to do this as it's all generic code, but
# just for extra caution we do it for each backend.
#

try:
    import trio

    @pytest.mark.skipif(not _nanonameserver_available, reason="requires nanonameserver")
    def test_trio_inbound_xfr():
        dns.asyncbackend.set_default_backend("trio")

        async def run():
            await async_inbound_xfr()

        trio.run(run)

except ImportError:
    pass


class UDPXFRNanoNameserver(Server):
    def __init__(self):
        super().__init__(origin=dns.name.from_text("example"))
        self.did_truncation = False

    def handle(self, request):
        try:
            if request.message.question[0].rdtype == dns.rdatatype.IXFR:
                if self.did_truncation:
                    text = ixfr
                else:
                    text = retry_tcp_ixfr
                    self.did_truncation = True
            else:
                text = axfr
            r = dns.message.from_text(text, one_rr_per_rrset=True, origin=self.origin)
            r.id = request.message.id
            return r
        except Exception:
            pass


@pytest.mark.skipif(not _nanonameserver_available, reason="requires nanonameserver")
def test_sync_retry_tcp_inbound_xfr():
    with UDPXFRNanoNameserver() as ns:
        zone = dns.versioned.Zone("example")
        dns.query.inbound_xfr(
            ns.tcp_address[0],
            zone,
            port=ns.tcp_address[1],
            udp_mode=dns.query.UDPMode.TRY_FIRST,
        )
        dns.query.inbound_xfr(
            ns.tcp_address[0],
            zone,
            port=ns.tcp_address[1],
            udp_mode=dns.query.UDPMode.TRY_FIRST,
        )
        expected = dns.zone.from_text(ixfr_expected, "example")
        assert zone == expected


async def udp_async_inbound_xfr():
    with UDPXFRNanoNameserver() as ns:
        zone = dns.versioned.Zone("example")
        await dns.asyncquery.inbound_xfr(
            ns.tcp_address[0],
            zone,
            port=ns.tcp_address[1],
            udp_mode=dns.query.UDPMode.TRY_FIRST,
        )
        await dns.asyncquery.inbound_xfr(
            ns.tcp_address[0],
            zone,
            port=ns.tcp_address[1],
            udp_mode=dns.query.UDPMode.TRY_FIRST,
        )
        expected = dns.zone.from_text(ixfr_expected, "example")
        assert zone == expected


@pytest.mark.skipif(not _nanonameserver_available, reason="requires nanonameserver")
def test_asyncio_retry_tcp_inbound_xfr():
    dns.asyncbackend.set_default_backend("asyncio")

    async def run():
        await udp_async_inbound_xfr()

    try:
        runner = asyncio.run
    except AttributeError:

        def old_runner(awaitable):
            loop = asyncio.get_event_loop()
            return loop.run_until_complete(awaitable)

        runner = old_runner
    runner(run())



# Copyright (C) Dnspython Contributors, see LICENSE for text of ISC license

import unittest

import dns.flags
import dns.message
import dns.name
import dns.rcode
import dns.rdataclass
import dns.rdatatype
import dns.resolver
import dns.tsigkeyring

# Test the resolver's Resolution, i.e. the business logic of the resolver.


class ResolutionTestCase(unittest.TestCase):
    def setUp(self):
        self.resolver = dns.resolver.Resolver(configure=False)
        self.resolver.nameservers = ["10.0.0.1", "10.0.0.2"]
        self.resolver.domain = dns.name.from_text("example")
        self.qname = dns.name.from_text("www.dnspython.org")
        self.resn = dns.resolver._Resolution(
            self.resolver, self.qname, "A", "IN", False, True, False
        )

    def test_next_request_abs(self):
        (request, answer) = self.resn.next_request()
        self.assertTrue(answer is None)
        self.assertEqual(request.question[0].name, self.qname)
        self.assertEqual(request.question[0].rdtype, dns.rdatatype.A)

    def test_next_request_rel_with_search(self):
        qname = dns.name.from_text("www.dnspython.org", None)
        abs_qname_1 = dns.name.from_text("www.dnspython.org.example")
        self.resn = dns.resolver._Resolution(
            self.resolver, qname, "A", "IN", False, True, True
        )
        (request, answer) = self.resn.next_request()
        self.assertTrue(answer is None)
        self.assertEqual(request.question[0].name, self.qname)
        self.assertEqual(request.question[0].rdtype, dns.rdatatype.A)
        (request, answer) = self.resn.next_request()
        self.assertTrue(answer is None)
        self.assertEqual(request.question[0].name, abs_qname_1)
        self.assertEqual(request.question[0].rdtype, dns.rdatatype.A)

        def bad():
            (request, answer) = self.resn.next_request()

        self.assertRaises(dns.resolver.NXDOMAIN, bad)

    def test_next_request_rel_without_search(self):
        qname = dns.name.from_text("www.dnspython.org", None)
        abs_qname_1 = dns.name.from_text("www.dnspython.org.example")
        self.resn = dns.resolver._Resolution(
            self.resolver, qname, "A", "IN", False, True, False
        )
        (request, answer) = self.resn.next_request()
        self.assertTrue(answer is None)
        self.assertEqual(request.question[0].name, self.qname)
        self.assertEqual(request.question[0].rdtype, dns.rdatatype.A)

        def bad():
            (request, answer) = self.resn.next_request()

        self.assertRaises(dns.resolver.NXDOMAIN, bad)

    def test_next_request_exhaust_causes_nxdomain(self):
        def bad():
            (request, answer) = self.resn.next_request()

        (request, answer) = self.resn.next_request()
        self.assertRaises(dns.resolver.NXDOMAIN, bad)

    def make_address_response(self, q):
        r = dns.message.make_response(q)
        rrs = r.get_rrset(
            r.answer, self.qname, dns.rdataclass.IN, dns.rdatatype.A, create=True
        )
        rrs.add(
            dns.rdata.from_text(dns.rdataclass.IN, dns.rdatatype.A, "10.0.0.1"), 300
        )
        return r

    def make_negative_response(self, q, nxdomain=False):
        r = dns.message.make_response(q)
        rrs = r.get_rrset(
            r.authority,
            q.question[0].name,
            dns.rdataclass.IN,
            dns.rdatatype.SOA,
            create=True,
        )
        rrs.add(
            dns.rdata.from_text(
                dns.rdataclass.IN, dns.rdatatype.SOA, ". . 1 2 3 4 300"
            ),
            300,
        )
        if nxdomain:
            r.set_rcode(dns.rcode.NXDOMAIN)
        return r

    def make_long_chain_response(self, q, count):
        r = dns.message.make_response(q)
        name = self.qname
        for i in range(count):
            rrs = r.get_rrset(
                r.answer, name, dns.rdataclass.IN, dns.rdatatype.CNAME, create=True
            )
            tname = dns.name.from_text(f"target{i}.")
            rrs.add(
                dns.rdata.from_text(dns.rdataclass.IN, dns.rdatatype.CNAME, str(tname)),
                300,
            )
            name = tname
        rrs = r.get_rrset(
            r.answer, name, dns.rdataclass.IN, dns.rdatatype.A, create=True
        )
        rrs.add(
            dns.rdata.from_text(dns.rdataclass.IN, dns.rdatatype.A, "10.0.0.1"), 300
        )
        return r

    def test_next_request_cache_hit(self):
        self.resolver.cache = dns.resolver.Cache()
        q = dns.message.make_query(self.qname, dns.rdatatype.A)
        r = self.make_address_response(q)
        cache_answer = dns.resolver.Answer(
            self.qname, dns.rdatatype.A, dns.rdataclass.IN, r
        )
        self.resolver.cache.put(
            (self.qname, dns.rdatatype.A, dns.rdataclass.IN), cache_answer
        )
        (request, answer) = self.resn.next_request()
        self.assertTrue(request is None)
        self.assertTrue(answer is cache_answer)

    def test_next_request_cached_no_answer(self):
        # In default mode, we should raise on a no-answer hit
        self.resolver.cache = dns.resolver.Cache()
        q = dns.message.make_query(self.qname, dns.rdatatype.A)
        # Note we need an SOA so the cache doesn't expire the answer
        # immediately, but our negative response code does that.
        r = self.make_negative_response(q)
        cache_answer = dns.resolver.Answer(
            self.qname, dns.rdatatype.A, dns.rdataclass.IN, r
        )
        self.resolver.cache.put(
            (self.qname, dns.rdatatype.A, dns.rdataclass.IN), cache_answer
        )

        def bad():
            (request, answer) = self.resn.next_request()

        self.assertRaises(dns.resolver.NoAnswer, bad)
        # If raise_on_no_answer is False, we should get a cache hit.
        self.resn = dns.resolver._Resolution(
            self.resolver, self.qname, "A", "IN", False, False, False
        )
        (request, answer) = self.resn.next_request()
        self.assertTrue(request is None)
        self.assertTrue(answer is cache_answer)

    def test_next_request_cached_nxdomain_without_search(self):
        # use a relative qname
        qname = dns.name.from_text("www.dnspython.org", None)
        self.resn = dns.resolver._Resolution(
            self.resolver, qname, "A", "IN", False, True, False
        )
        qname1 = dns.name.from_text("www.dnspython.org.")
        # Arrange to get NXDOMAIN hits on it.
        self.resolver.cache = dns.resolver.Cache()
        q1 = dns.message.make_query(qname1, dns.rdatatype.A)
        r1 = self.make_negative_response(q1, True)
        cache_answer = dns.resolver.Answer(
            qname1, dns.rdatatype.ANY, dns.rdataclass.IN, r1
        )
        self.resolver.cache.put(
            (qname1, dns.rdatatype.ANY, dns.rdataclass.IN), cache_answer
        )
        try:
            (request, answer) = self.resn.next_request()
            self.assertTrue(False)  # should not happen!
        except dns.resolver.NXDOMAIN as nx:
            self.assertTrue(nx.response(qname1) is r1)

    def test_next_request_cached_nxdomain_with_search(self):
        # use a relative qname so we have two qnames to try
        qname = dns.name.from_text("www.dnspython.org", None)
        # also enable search mode or we'll only see www.dnspython.org.
        self.resn = dns.resolver._Resolution(
            self.resolver, qname, "A", "IN", False, True, True
        )
        qname1 = dns.name.from_text("www.dnspython.org.example.")
        qname2 = dns.name.from_text("www.dnspython.org.")
        # Arrange to get NXDOMAIN hits on both of those qnames.
        self.resolver.cache = dns.resolver.Cache()
        q1 = dns.message.make_query(qname1, dns.rdatatype.A)
        r1 = self.make_negative_response(q1, True)
        cache_answer = dns.resolver.Answer(
            qname1, dns.rdatatype.ANY, dns.rdataclass.IN, r1
        )
        self.resolver.cache.put(
            (qname1, dns.rdatatype.ANY, dns.rdataclass.IN), cache_answer
        )
        q2 = dns.message.make_query(qname2, dns.rdatatype.A)
        r2 = self.make_negative_response(q2, True)
        cache_answer = dns.resolver.Answer(
            qname2, dns.rdatatype.ANY, dns.rdataclass.IN, r2
        )
        self.resolver.cache.put(
            (qname2, dns.rdatatype.ANY, dns.rdataclass.IN), cache_answer
        )
        try:
            (request, answer) = self.resn.next_request()
            self.assertTrue(False)  # should not happen!
        except dns.resolver.NXDOMAIN as nx:
            self.assertTrue(nx.response(qname1) is r1)
            self.assertTrue(nx.response(qname2) is r2)

    def test_next_request_rotate(self):
        self.resolver.rotate = True
        order1 = ["Do53:10.0.0.1@53", "Do53:10.0.0.2@53"]
        order2 = ["Do53:10.0.0.2@53", "Do53:10.0.0.1@53"]
        seen1 = False
        seen2 = False
        # We're not interested in testing the randomness, but we'd
        # like to see some shuffling, so try up to 50 times to see
        # both orders at least once.  This test can fail even with
        # correct code, but it is *extremely* unlikely.
        for count in range(0, 50):
            self.resn = dns.resolver._Resolution(
                self.resolver, self.qname, "A", "IN", False, True, False
            )
            self.resn.next_request()
            text_form = [str(n) for n in self.resn.nameservers]
            print(text_form)
            if text_form == order1:
                seen1 = True
            elif text_form == order2:
                seen2 = True
            else:
                raise ValueError  # should not happen!
            if seen1 and seen2:
                break
        self.assertTrue(seen1 and seen2)

    def test_next_request_TSIG(self):
        self.resolver.keyring = dns.tsigkeyring.from_text(
            {"keyname.": "NjHwPsMKjdN++dOfE5iAiQ=="}
        )
        (keyname, secret) = next(iter(self.resolver.keyring.items()))
        self.resolver.keyname = dns.name.from_text("keyname.")
        (request, answer) = self.resn.next_request()
        self.assertFalse(request is None)
        self.assertEqual(request.keyring.name, keyname)
        self.assertEqual(request.keyring.secret, secret)

    def test_next_request_flags(self):
        self.resolver.flags = dns.flags.RD | dns.flags.CD
        (request, answer) = self.resn.next_request()
        self.assertFalse(request is None)
        self.assertEqual(request.flags, self.resolver.flags)

    def test_next_nameserver_udp(self):
        (request, answer) = self.resn.next_request()
        (nameserver1, tcp, backoff) = self.resn.next_nameserver()
        self.assertEqual(nameserver1.port, 53)
        self.assertFalse(tcp)
        self.assertEqual(backoff, 0.0)
        (nameserver2, tcp, backoff) = self.resn.next_nameserver()
        self.assertTrue(nameserver2 != nameserver1)
        self.assertEqual(nameserver2.port, 53)
        self.assertFalse(tcp)
        self.assertEqual(backoff, 0.0)
        (nameserver3, tcp, backoff) = self.resn.next_nameserver()
        self.assertTrue(nameserver3 is nameserver1)
        self.assertFalse(tcp)
        self.assertEqual(backoff, 0.1)
        (nameserver4, tcp, backoff) = self.resn.next_nameserver()
        self.assertTrue(nameserver4 is nameserver2)
        self.assertFalse(tcp)
        self.assertEqual(backoff, 0.0)
        (nameserver5, tcp, backoff) = self.resn.next_nameserver()
        self.assertTrue(nameserver5 is nameserver1)
        self.assertFalse(tcp)
        self.assertEqual(backoff, 0.2)

    def test_next_nameserver_retry_with_tcp(self):
        (request, answer) = self.resn.next_request()
        (nameserver1, tcp, backoff) = self.resn.next_nameserver()
        self.assertEqual(nameserver1.port, 53)
        self.assertFalse(tcp)
        self.assertEqual(backoff, 0.0)
        self.resn.retry_with_tcp = True
        (nameserver2, tcp, backoff) = self.resn.next_nameserver()
        self.assertTrue(nameserver2 is nameserver1)
        self.assertTrue(tcp)
        self.assertEqual(backoff, 0.0)
        (nameserver3, tcp, backoff) = self.resn.next_nameserver()
        self.assertTrue(nameserver3 != nameserver1)
        self.assertEqual(nameserver3.port, 53)
        self.assertFalse(tcp)
        self.assertEqual(backoff, 0.0)

    def test_next_nameserver_no_nameservers(self):
        (request, answer) = self.resn.next_request()
        (nameserver, _, _) = self.resn.next_nameserver()
        self.resn.nameservers.remove(nameserver)
        (nameserver, _, _) = self.resn.next_nameserver()
        self.resn.nameservers.remove(nameserver)

        def bad():
            (nameserver, _, _) = self.resn.next_nameserver()

        self.assertRaises(dns.resolver.NoNameservers, bad)

    def test_next_nameserver_max_size_nameserver(self):
        # A query to a nameserver that always supports a maximum size query
        # always counts as a "tcp attempt" for the state machine
        self.resolver.nameservers = ["https://127.0.0.1:443/bogus"]
        (_, _) = self.resn.next_request()
        (nameserver, tcp_attempt, _) = self.resn.next_nameserver()
        print(nameserver)
        assert tcp_attempt

    def test_query_result_nameserver_removing_exceptions(self):
        # add some nameservers so we have enough to remove :)
        new_nameservers = list(self.resolver.nameservers[:])
        new_nameservers.extend(["10.0.0.3", "10.0.0.4"])
        self.resolver.nameservers = new_nameservers
        (request, _) = self.resn.next_request()
        exceptions = [
            dns.exception.FormError(),
            EOFError(),
            NotImplementedError(),
            dns.message.Truncated(),
        ]
        for i in range(4):
            (nameserver, _, _) = self.resn.next_nameserver()
            if i == 3:
                # Truncated is only bad if we're doing TCP, make it look
                # like that's the case
                self.resn.tcp_attempt = True
            self.assertTrue(nameserver in self.resn.nameservers)
            (answer, done) = self.resn.query_result(None, exceptions[i])
            self.assertTrue(answer is None)
            self.assertFalse(done)
            self.assertFalse(nameserver in self.resn.nameservers)
        self.assertEqual(len(self.resn.nameservers), 0)

    def test_query_result_nameserver_continuing_exception(self):
        # except for the exceptions tested in
        # test_query_result_nameserver_removing_exceptions(), we should
        # not remove any nameservers and just continue resolving.
        (_, _) = self.resn.next_request()
        (_, _, _) = self.resn.next_nameserver()
        nameservers = self.resn.nameservers[:]
        (answer, done) = self.resn.query_result(None, dns.exception.Timeout())
        self.assertTrue(answer is None)
        self.assertFalse(done)
        self.assertEqual(nameservers, self.resn.nameservers)

    def test_query_result_retry_with_tcp(self):
        (request, _) = self.resn.next_request()
        (nameserver, tcp, _) = self.resn.next_nameserver()
        self.assertFalse(tcp)
        (answer, done) = self.resn.query_result(None, dns.message.Truncated())
        self.assertTrue(answer is None)
        self.assertFalse(done)
        self.assertTrue(self.resn.retry_with_tcp)
        # The rest of TCP retry logic was tested above in
        # test_next_nameserver_retry_with_tcp(), so we do not repeat
        # it.

    def test_query_result_no_error_with_data(self):
        q = dns.message.make_query(self.qname, dns.rdatatype.A)
        r = self.make_address_response(q)
        (_, _) = self.resn.next_request()
        (_, _, _) = self.resn.next_nameserver()
        (answer, done) = self.resn.query_result(r, None)
        self.assertFalse(answer is None)
        self.assertTrue(done)
        self.assertEqual(answer.qname, self.qname)
        self.assertEqual(answer.rdtype, dns.rdatatype.A)

    def test_query_result_no_error_with_data_cached(self):
        self.resolver.cache = dns.resolver.Cache()
        q = dns.message.make_query(self.qname, dns.rdatatype.A)
        r = self.make_address_response(q)
        (_, _) = self.resn.next_request()
        (_, _, _) = self.resn.next_nameserver()
        (answer, done) = self.resn.query_result(r, None)
        self.assertFalse(answer is None)
        cache_answer = self.resolver.cache.get(
            (self.qname, dns.rdatatype.A, dns.rdataclass.IN)
        )
        self.assertTrue(answer is cache_answer)

    def test_query_result_no_error_no_data(self):
        q = dns.message.make_query(self.qname, dns.rdatatype.A)
        r = self.make_negative_response(q)
        (_, _) = self.resn.next_request()
        (_, _, _) = self.resn.next_nameserver()

        def bad():
            (answer, done) = self.resn.query_result(r, None)

        self.assertRaises(dns.resolver.NoAnswer, bad)

    def test_query_result_nxdomain(self):
        q = dns.message.make_query(self.qname, dns.rdatatype.A)
        r = self.make_negative_response(q, True)
        (_, _) = self.resn.next_request()
        (_, _, _) = self.resn.next_nameserver()
        (answer, done) = self.resn.query_result(r, None)
        self.assertTrue(answer is None)
        self.assertTrue(done)

    def test_query_result_nxdomain_but_has_answer(self):
        q = dns.message.make_query(self.qname, dns.rdatatype.A)
        r = self.make_address_response(q)
        r.set_rcode(dns.rcode.NXDOMAIN)
        (_, _) = self.resn.next_request()
        (nameserver, _, _) = self.resn.next_nameserver()
        (answer, done) = self.resn.query_result(r, None)
        self.assertIsNone(answer)
        self.assertFalse(done)
        self.assertTrue(nameserver not in self.resn.nameservers)

    def test_query_result_chain_not_too_long(self):
        q = dns.message.make_query(self.qname, dns.rdatatype.A)
        r = self.make_long_chain_response(q, 15)
        (_, _) = self.resn.next_request()
        (_, _, _) = self.resn.next_nameserver()
        (answer, done) = self.resn.query_result(r, None)
        self.assertIsNotNone(answer)
        self.assertTrue(done)

    def test_query_result_chain_too_long(self):
        q = dns.message.make_query(self.qname, dns.rdatatype.A)
        r = self.make_long_chain_response(q, 16)
        (_, _) = self.resn.next_request()
        (nameserver, _, _) = self.resn.next_nameserver()
        (answer, done) = self.resn.query_result(r, None)
        self.assertIsNone(answer)
        self.assertFalse(done)
        self.assertTrue(nameserver not in self.resn.nameservers)

    def test_query_result_nxdomain_cached(self):
        self.resolver.cache = dns.resolver.Cache()
        q = dns.message.make_query(self.qname, dns.rdatatype.A)
        r = self.make_negative_response(q, True)
        (_, _) = self.resn.next_request()
        (_, _, _) = self.resn.next_nameserver()
        (answer, done) = self.resn.query_result(r, None)
        self.assertTrue(answer is None)
        self.assertTrue(done)
        cache_answer = self.resolver.cache.get(
            (self.qname, dns.rdatatype.ANY, dns.rdataclass.IN)
        )
        self.assertTrue(cache_answer.response is r)

    def test_query_result_yxdomain(self):
        q = dns.message.make_query(self.qname, dns.rdatatype.A)
        r = self.make_address_response(q)
        r.set_rcode(dns.rcode.YXDOMAIN)
        (_, _) = self.resn.next_request()
        (_, _, _) = self.resn.next_nameserver()

        def bad():
            (answer, done) = self.resn.query_result(r, None)

        self.assertRaises(dns.resolver.YXDOMAIN, bad)

    def test_query_result_servfail_no_retry(self):
        q = dns.message.make_query(self.qname, dns.rdatatype.A)
        r = self.make_address_response(q)
        r.set_rcode(dns.rcode.SERVFAIL)
        (_, _) = self.resn.next_request()
        (nameserver, _, _) = self.resn.next_nameserver()
        (answer, done) = self.resn.query_result(r, None)
        self.assertTrue(answer is None)
        self.assertFalse(done)
        self.assertTrue(nameserver not in self.resn.nameservers)

    def test_query_result_servfail_with_retry(self):
        self.resolver.retry_servfail = True
        q = dns.message.make_query(self.qname, dns.rdatatype.A)
        r = self.make_address_response(q)
        r.set_rcode(dns.rcode.SERVFAIL)
        (_, _) = self.resn.next_request()
        (_, _, _) = self.resn.next_nameserver()
        nameservers = self.resn.nameservers[:]
        (answer, done) = self.resn.query_result(r, None)
        self.assertTrue(answer is None)
        self.assertFalse(done)
        self.assertEqual(nameservers, self.resn.nameservers)

    def test_query_result_other_unhappy_rcode(self):
        q = dns.message.make_query(self.qname, dns.rdatatype.A)
        r = self.make_address_response(q)
        r.set_rcode(dns.rcode.REFUSED)
        (_, _) = self.resn.next_request()
        (nameserver, _, _) = self.resn.next_nameserver()
        (answer, done) = self.resn.query_result(r, None)
        self.assertTrue(answer is None)
        self.assertFalse(done)
        self.assertTrue(nameserver not in self.resn.nameservers)

    def test_no_metaqueries(self):
        def bad1():
            self.resn = dns.resolver._Resolution(
                self.resolver, self.qname, "ANY", "IN", False, True, False
            )

        def bad2():
            self.resn = dns.resolver._Resolution(
                self.resolver, self.qname, "A", "ANY", False, True, False
            )

        self.assertRaises(dns.resolver.NoMetaqueries, bad1)
        self.assertRaises(dns.resolver.NoMetaqueries, bad2)



# Copyright (C) Dnspython Contributors, see LICENSE for text of ISC license

import pytest

from dns._features import (
    _cache,
    _requirements,
    _tuple_from_text,
    _version_check,
    force,
    have,
)

try:
    import cryptography

    v = _tuple_from_text(cryptography.__version__)
    have_cryptography = v >= (42, 0, 0)
except ImportError:
    have_cryptography = False


def test_tuple_from_text():
    assert _tuple_from_text("") == ()
    assert _tuple_from_text("1") == (1,)
    assert _tuple_from_text("1.2") == (1, 2)
    assert _tuple_from_text("1.2rc1") == (1, 2)
    assert _tuple_from_text("1.2.junk3") == (1, 2)


@pytest.mark.skipif(
    not have_cryptography, reason="cryptography not available or too old"
)
def test_version_check():
    assert _version_check("cryptography>=42")
    assert not _version_check("cryptography>=10000")
    assert not _version_check("totallyboguspackagename>=10000")


@pytest.mark.skipif(
    not have_cryptography, reason="cryptography not available or too old"
)
def test_have():
    # ensure cache is empty; we can't just assign as our local is shadowing the
    # variable in dns._features
    while len(_cache) > 0:
        _cache.popitem()
    assert have("dnssec")
    assert _cache["dnssec"] == True
    assert not have("bogusfeature")
    assert _cache["bogusfeature"] == False
    _requirements["unavailable"] = ["bogusmodule>=10000"]
    try:
        assert not have("unavailable")
    finally:
        del _requirements["unavailable"]


def test_force():
    while len(_cache) > 0:
        _cache.popitem()
    assert not have("bogusfeature")
    assert _cache["bogusfeature"] == False
    force("bogusfeature", True)
    assert have("bogusfeature")
    assert _cache["bogusfeature"] == True
    force("bogusfeature", False)
    assert not have("bogusfeature")
    assert _cache["bogusfeature"] == False
    _requirements["unavailable"] = ["bogusmodule>=10000"]
    try:
        assert not have("unavailable")
        assert _cache["unavailable"] == False
        force("unavailable", True)
        assert _cache["unavailable"] == True
    finally:
        del _requirements["unavailable"]



#!/usr/bin/env python3
import dns.edns
import dns.message
import dns.query

# This example demonstrates how to use the EDNS client subnet option

ADDRESS = "0.0.0.0"  # replace this with the address you want to check
PREFIX = 0  # replace this with a prefix length (typically 24 for IPv4)

ecs = dns.edns.ECSOption(ADDRESS, PREFIX)
q = dns.message.make_query("www.google.com", "A", use_edns=0, options=[ecs])
r = dns.query.udp(q, "8.8.8.8")
print(r)



#!/usr/bin/env python3

import dns.edns
import dns.message
import dns.query
import dns.resolver

n = "."
t = dns.rdatatype.SOA
l = "199.7.83.42"  # Address of l.root-servers.net
i = "149.20.1.73"  # Address of ns1.isc.org, for COOKIEs

q_list = []

# A query without EDNS0
q_list.append((l, dns.message.make_query(n, t)))

# The same query, but with EDNS0 turned on with no options
q_list.append((l, dns.message.make_query(n, t, use_edns=0)))

# Use use_edns() to specify EDNS0 options, such as buffer size
this_q = dns.message.make_query(n, t)
this_q.use_edns(0, payload=2000)
q_list.append((l, this_q))

# With an NSID option
#   use_edns=0 is not needed if options are specified)
q_list.append(
    (
        l,
        dns.message.make_query(
            n, t, options=[dns.edns.GenericOption(dns.edns.OptionType.NSID, b"")]
        ),
    )
)

# With an NSID option, but with use_edns() to specify the options
this_q = dns.message.make_query(n, t)
this_q.use_edns(0, options=[dns.edns.GenericOption(dns.edns.OptionType.NSID, b"")])
q_list.append((l, this_q))

# With a COOKIE
q_list.append(
    (
        i,
        dns.message.make_query(
            n,
            t,
            options=[
                dns.edns.GenericOption(
                    dns.edns.OptionType.COOKIE, b"0xfe11ac99bebe3322"
                )
            ],
        ),
    )
)

# With an ECS option using dns.edns.ECSOption to form the option
q_list.append(
    (l, dns.message.make_query(n, t, options=[dns.edns.ECSOption("192.168.0.0", 20)]))
)

for addr, q in q_list:
    r = dns.query.udp(q, addr)
    if not r.options:
        print("No EDNS options returned")
    else:
        for o in r.options:
            print(o.otype.value, o.data)
    print()



# Copyright (C) Dnspython Contributors, see LICENSE for text of ISC license

# Copyright (C) 2003-2007, 2009-2011 Nominum, Inc.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose with or without fee is hereby granted,
# provided that the above copyright notice and this permission notice
# appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NOMINUM DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NOMINUM BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import enum
import inspect
import os

import dns.message
import dns.name
import dns.query
import dns.rdataclass
import dns.rdatatype

# Cache for is_internet_reachable()
_internet_reachable = None
_have_ipv4 = False
_have_ipv6 = False


def here(filename):
    return os.path.join(os.path.dirname(__file__), filename)


def check_networking(addresses):
    """Can we do a DNS resolution via UDP and TCP to at least one of the addresses?"""
    for address in addresses:
        try:
            q = dns.message.make_query(dns.name.root, dns.rdatatype.NS)
            ok = False
            # We try UDP a few times in case we get unlucky and a packet is lost.
            for i in range(5):
                # We don't check the answer other than make sure there is one.
                try:
                    r = dns.query.udp(q, address, timeout=4)
                    ns = r.find_rrset(
                        r.answer, dns.name.root, dns.rdataclass.IN, dns.rdatatype.NS
                    )
                    ok = True
                    break
                except Exception:
                    continue  # UDP try loop
            if not ok:
                continue  # addresses loop
            try:
                r = dns.query.tcp(q, address, timeout=4)
                ns = r.find_rrset(
                    r.answer, dns.name.root, dns.rdataclass.IN, dns.rdatatype.NS
                )
                # UDP and TCP both work!
                return True
            except Exception:
                continue
        except Exception as e:
            pass
    return False


def is_internet_reachable():
    """Check if the Internet is reachable.

    Setting the environment variable `NO_INTERNET` will let this
    function always return False. The result is cached.

    We check using the Google and Cloudflare public resolvers as they are highly
    available and have well-known stable addresses.
    """
    global _internet_reachable
    if _internet_reachable is None:
        if os.environ.get("NO_INTERNET"):
            _internet_reachable = False
        else:
            global _have_ipv4
            _have_ipv4 = check_networking(["8.8.8.8", "1.1.1.1"])
            global _have_ipv6
            _have_ipv6 = check_networking(
                ["2001:4860:4860::8888", "2606:4700:4700::1111"]
            )
            _internet_reachable = _have_ipv4 or _have_ipv6
    return _internet_reachable


def have_ipv4():
    if not is_internet_reachable():
        return False
    return _have_ipv4


def have_ipv6():
    if not is_internet_reachable():
        return False
    return _have_ipv6


def enumerate_module(module, super_class):
    """Yield module attributes which are subclasses of given class"""
    for attr_name in dir(module):
        attr = getattr(module, attr_name)
        if inspect.isclass(attr) and issubclass(attr, super_class):
            yield attr


def check_enum_exports(module, eq_callback, only=None):
    """Make sure module exports all mnemonics from enums"""
    for attr in enumerate_module(module, enum.Enum):
        if only is not None and attr not in only:
            # print('SKIP', attr)
            continue
        for flag, value in attr.__members__.items():
            # print(module, flag, value)
            eq_callback(getattr(module, flag), value)


def is_docker() -> bool:
    # There are a lot of answers to "am I runnning in a container" and none appear
    # to work reliably, so we're settling for "am I running in docker?"
    try:
        return os.path.isfile("/.dockerenv")
    except Exception:
        return False



import asyncio
import threading

import dns.asyncbackend
import dns.asyncquery
import dns.message
import dns.query
import dns.quic
import dns.rdatatype

try:
    import trio

    have_trio = True
except ImportError:
    have_trio = False

# This demo assumes you have the aioquic example doq_server.py running on localhost
# on port 4784 on localhost.
peer_address = "127.0.0.1"
peer_port = 4784
query_name = "www.dnspython.org"
tls_verify_mode = False


def squery(rdtype="A", connection=None):
    q = dns.message.make_query(query_name, rdtype)
    r = dns.query.quic(
        q, peer_address, port=peer_port, connection=connection, verify=tls_verify_mode
    )
    print(r)


def srun():
    squery()


def smultirun():
    with dns.quic.SyncQuicManager(verify_mode=tls_verify_mode) as manager:
        connection = manager.connect(peer_address, peer_port)
        t1 = threading.Thread(target=squery, args=["A", connection])
        t1.start()
        t2 = threading.Thread(target=squery, args=["AAAA", connection])
        t2.start()
        t1.join()
        t2.join()


async def aquery(rdtype="A", connection=None):
    q = dns.message.make_query(query_name, rdtype)
    r = await dns.asyncquery.quic(
        q, peer_address, port=peer_port, connection=connection, verify=tls_verify_mode
    )
    print(r)


def arun():
    asyncio.run(aquery())


async def amulti():
    async with dns.quic.AsyncioQuicManager(verify_mode=tls_verify_mode) as manager:
        connection = manager.connect(peer_address, peer_port)
        t1 = asyncio.Task(aquery("A", connection))
        t2 = asyncio.Task(aquery("AAAA", connection))
        await t1
        await t2


def amultirun():
    asyncio.run(amulti())


if have_trio:

    def trun():
        trio.run(aquery)

    async def tmulti():
        async with trio.open_nursery() as nursery:
            async with dns.quic.TrioQuicManager(
                nursery, verify_mode=tls_verify_mode
            ) as manager:
                async with trio.open_nursery() as query_nursery:
                    # We run queries in a separate nursery so we can demonstrate
                    # waiting for them all to exit without waiting for the manager to
                    # exit as well.
                    connection = manager.connect(peer_address, peer_port)
                    query_nursery.start_soon(aquery, "A", connection)
                    query_nursery.start_soon(aquery, "AAAA", connection)

    def tmultirun():
        trio.run(tmulti)


def main():
    print("*** Single Queries ***")
    print("--- Sync ---")
    srun()
    print("--- Asyncio ---")
    dns.asyncbackend.set_default_backend("asyncio")
    arun()
    if have_trio:
        print("--- Trio ---")
        dns.asyncbackend.set_default_backend("trio")
        trun()
    print("*** Multi-connection Queries ***")
    print("--- Sync ---")
    smultirun()
    print("--- Asyncio ---")
    dns.asyncbackend.set_default_backend("asyncio")
    amultirun()
    if have_trio:
        print("--- Trio ---")
        dns.asyncbackend.set_default_backend("trio")
        tmultirun()


if __name__ == "__main__":
    main()



# Copyright (C) Dnspython Contributors, see LICENSE for text of ISC license

# Copyright (C) 2003-2017 Nominum, Inc.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose with or without fee is hereby granted,
# provided that the above copyright notice and this permission notice
# appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NOMINUM DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NOMINUM BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import asyncio
import random
import socket
import time
import unittest

import dns.asyncbackend
import dns.asyncquery
import dns.asyncresolver
import dns.message
import dns.name
import dns.query
import dns.rcode
import dns.rdataclass
import dns.rdatatype
import dns.resolver
import tests.util

# Some tests require TLS so skip those if it's not there.
ssl = dns.query.ssl
try:
    ssl.create_default_context()
    _ssl_available = True
except Exception:
    _ssl_available = False


# Look for systemd-resolved, as it does dangling CNAME responses incorrectly.
#
# Currently we simply check if the nameserver is 127.0.0.53.
_systemd_resolved_present = False
try:
    _resolver = dns.resolver.Resolver()
    if _resolver.nameservers == ["127.0.0.53"]:
        _systemd_resolved_present = True
except Exception:
    pass

# Docker too!  It not only has problems with dangling CNAME, but also says NXDOMAIN when
# it should say no error no data.
_is_docker = tests.util.is_docker()

query_addresses = []
family = socket.AF_UNSPEC
if tests.util.have_ipv4():
    query_addresses.append("8.8.8.8")
    family = socket.AF_INET
if tests.util.have_ipv6():
    have_v6 = True
    if family == socket.AF_INET:
        # we have both working, go back to UNSPEC
        family = socket.AF_UNSPEC
    else:
        # v6 only
        family = socket.AF_INET6
    query_addresses.append("2001:4860:4860::8888")

KNOWN_ANYCAST_DOH_RESOLVER_URLS = [
    "https://cloudflare-dns.com/dns-query",
    "https://dns.google/dns-query",
    # 'https://dns11.quad9.net/dns-query',
]


class AsyncDetectionTests(unittest.TestCase):
    sniff_result = "asyncio"

    def async_run(self, afunc):
        return asyncio.run(afunc())

    def test_sniff(self):
        dns.asyncbackend._default_backend = None

        async def run():
            self.assertEqual(dns.asyncbackend.sniff(), self.sniff_result)

        self.async_run(run)

    def test_get_default_backend(self):
        dns.asyncbackend._default_backend = None

        async def run():
            backend = dns.asyncbackend.get_default_backend()
            self.assertEqual(backend.name(), self.sniff_result)

        self.async_run(run)


class NoSniffioAsyncDetectionTests(AsyncDetectionTests):
    expect_raise = False

    def setUp(self):
        dns.asyncbackend._no_sniffio = True

    def tearDown(self):
        dns.asyncbackend._no_sniffio = False

    def test_sniff(self):
        dns.asyncbackend._default_backend = None
        if self.expect_raise:

            async def abad():
                dns.asyncbackend.sniff()

            def bad():
                self.async_run(abad)

            self.assertRaises(dns.asyncbackend.AsyncLibraryNotFoundError, bad)
        else:
            super().test_sniff()

    def test_get_default_backend(self):
        dns.asyncbackend._default_backend = None
        if self.expect_raise:

            async def abad():
                dns.asyncbackend.get_default_backend()

            def bad():
                self.async_run(abad)

            self.assertRaises(dns.asyncbackend.AsyncLibraryNotFoundError, bad)
        else:
            super().test_get_default_backend()


class MiscBackend(unittest.TestCase):
    def test_sniff_without_run_loop(self):
        dns.asyncbackend._default_backend = None

        def bad():
            dns.asyncbackend.sniff()

        self.assertRaises(dns.asyncbackend.AsyncLibraryNotFoundError, bad)

    def test_bogus_backend(self):
        def bad():
            dns.asyncbackend.get_backend("bogus")

        self.assertRaises(NotImplementedError, bad)


class MiscQuery(unittest.TestCase):
    def test_source_tuple(self):
        t = dns.asyncquery._source_tuple(socket.AF_INET, None, 0)
        self.assertEqual(t, None)
        t = dns.asyncquery._source_tuple(socket.AF_INET6, None, 0)
        self.assertEqual(t, None)
        t = dns.asyncquery._source_tuple(socket.AF_INET, "1.2.3.4", 53)
        self.assertEqual(t, ("1.2.3.4", 53))
        t = dns.asyncquery._source_tuple(socket.AF_INET6, "1::2", 53)
        self.assertEqual(t, ("1::2", 53))
        t = dns.asyncquery._source_tuple(socket.AF_INET, None, 53)
        self.assertEqual(t, ("0.0.0.0", 53))
        t = dns.asyncquery._source_tuple(socket.AF_INET6, None, 53)
        self.assertEqual(t, ("::", 53))


@unittest.skipIf(not tests.util.is_internet_reachable(), "Internet not reachable")
class AsyncTests(unittest.TestCase):
    def setUp(self):
        self.backend = dns.asyncbackend.set_default_backend("asyncio")

    def async_run(self, afunc):
        return asyncio.run(afunc())

    def testResolve(self):
        async def run():
            answer = await dns.asyncresolver.resolve("dns.google.", "A")
            return set([rdata.address for rdata in answer])

        seen = self.async_run(run)
        self.assertTrue("8.8.8.8" in seen)
        self.assertTrue("8.8.4.4" in seen)

    def testResolveAddress(self):
        async def run():
            return await dns.asyncresolver.resolve_address("8.8.8.8")

        answer = self.async_run(run)
        dnsgoogle = dns.name.from_text("dns.google.")
        self.assertEqual(answer[0].target, dnsgoogle)

    def testResolveName(self):
        async def run1():
            return await dns.asyncresolver.resolve_name("dns.google.")

        answers = self.async_run(run1)
        seen = set(answers.addresses())
        self.assertEqual(len(seen), 4)
        self.assertIn("8.8.8.8", seen)
        self.assertIn("8.8.4.4", seen)
        self.assertIn("2001:4860:4860::8844", seen)
        self.assertIn("2001:4860:4860::8888", seen)

        async def run2():
            return await dns.asyncresolver.resolve_name("dns.google.", socket.AF_INET)

        answers = self.async_run(run2)
        seen = set(answers.addresses())
        self.assertEqual(len(seen), 2)
        self.assertIn("8.8.8.8", seen)
        self.assertIn("8.8.4.4", seen)

        async def run3():
            return await dns.asyncresolver.resolve_name("dns.google.", socket.AF_INET6)

        answers = self.async_run(run3)
        seen = set(answers.addresses())
        self.assertEqual(len(seen), 2)
        self.assertIn("2001:4860:4860::8844", seen)
        self.assertIn("2001:4860:4860::8888", seen)

        async def run4():
            await dns.asyncresolver.resolve_name("nxdomain.dnspython.org")

        with self.assertRaises(dns.resolver.NXDOMAIN):
            self.async_run(run4)

        async def run5():
            await dns.asyncresolver.resolve_name(
                dns.reversename.from_address("8.8.8.8")
            )

        if not _is_docker:
            # docker returns NXDOMAIN!
            with self.assertRaises(dns.resolver.NoAnswer):
                self.async_run(run5)

    def testCanonicalNameNoCNAME(self):
        cname = dns.name.from_text("www.google.com")

        async def run():
            return await dns.asyncresolver.canonical_name("www.google.com")

        self.assertEqual(self.async_run(run), cname)

    def testCanonicalNameCNAME(self):
        name = dns.name.from_text("www.dnspython.org")
        cname = dns.name.from_text("dmfrjf4ips8xa.cloudfront.net")

        async def run():
            return await dns.asyncresolver.canonical_name(name)

        self.assertEqual(self.async_run(run), cname)

    @unittest.skipIf(
        _systemd_resolved_present or _is_docker, "systemd-resolved or docker in use"
    )
    def testCanonicalNameDangling(self):
        name = dns.name.from_text("dangling-cname.dnspython.org")
        cname = dns.name.from_text("dangling-target.dnspython.org")

        async def run():
            return await dns.asyncresolver.canonical_name(name)

        self.assertEqual(self.async_run(run), cname)

    def testZoneForName1(self):
        async def run():
            name = dns.name.from_text("www.dnspython.org.")
            return await dns.asyncresolver.zone_for_name(name)

        ezname = dns.name.from_text("dnspython.org.")
        zname = self.async_run(run)
        self.assertEqual(zname, ezname)

    def testZoneForName2(self):
        async def run():
            name = dns.name.from_text("a.b.www.dnspython.org.")
            return await dns.asyncresolver.zone_for_name(name)

        ezname = dns.name.from_text("dnspython.org.")
        zname = self.async_run(run)
        self.assertEqual(zname, ezname)

    def testZoneForName3(self):
        async def run():
            name = dns.name.from_text("dnspython.org.")
            return await dns.asyncresolver.zone_for_name(name)

        ezname = dns.name.from_text("dnspython.org.")
        zname = self.async_run(run)
        self.assertEqual(zname, ezname)

    def testZoneForName4(self):
        def bad():
            name = dns.name.from_text("dnspython.org", None)

            async def run():
                return await dns.asyncresolver.zone_for_name(name)

            self.async_run(run)

        self.assertRaises(dns.resolver.NotAbsolute, bad)

    def testQueryUDP(self):
        for address in query_addresses:
            qname = dns.name.from_text("dns.google.")

            async def run():
                q = dns.message.make_query(qname, dns.rdatatype.A)
                return await dns.asyncquery.udp(q, address, timeout=2)

            response = self.async_run(run)
            rrs = response.get_rrset(
                response.answer, qname, dns.rdataclass.IN, dns.rdatatype.A
            )
            self.assertTrue(rrs is not None)
            seen = set([rdata.address for rdata in rrs])
            self.assertTrue("8.8.8.8" in seen)
            self.assertTrue("8.8.4.4" in seen)

    def testQueryUDPWithSocket(self):
        for address in query_addresses:
            qname = dns.name.from_text("dns.google.")

            async def run():
                async with await self.backend.make_socket(
                    dns.inet.af_for_address(address),
                    socket.SOCK_DGRAM,
                    0,
                    None,
                    None,
                ) as s:
                    q = dns.message.make_query(qname, dns.rdatatype.A)
                    return await dns.asyncquery.udp(q, address, sock=s, timeout=2)

            response = self.async_run(run)
            rrs = response.get_rrset(
                response.answer, qname, dns.rdataclass.IN, dns.rdatatype.A
            )
            self.assertTrue(rrs is not None)
            seen = set([rdata.address for rdata in rrs])
            self.assertTrue("8.8.8.8" in seen)
            self.assertTrue("8.8.4.4" in seen)

    def testQueryTCP(self):
        for address in query_addresses:
            qname = dns.name.from_text("dns.google.")

            async def run():
                q = dns.message.make_query(qname, dns.rdatatype.A)
                return await dns.asyncquery.tcp(q, address, timeout=2)

            response = self.async_run(run)
            rrs = response.get_rrset(
                response.answer, qname, dns.rdataclass.IN, dns.rdatatype.A
            )
            self.assertTrue(rrs is not None)
            seen = set([rdata.address for rdata in rrs])
            self.assertTrue("8.8.8.8" in seen)
            self.assertTrue("8.8.4.4" in seen)

    def testQueryTCPWithSocket(self):
        for address in query_addresses:
            qname = dns.name.from_text("dns.google.")

            async def run():
                async with await self.backend.make_socket(
                    dns.inet.af_for_address(address),
                    socket.SOCK_STREAM,
                    0,
                    None,
                    (address, 53),
                    2,
                ) as s:
                    # for basic coverage
                    await s.getsockname()
                    q = dns.message.make_query(qname, dns.rdatatype.A)
                    return await dns.asyncquery.tcp(q, address, sock=s, timeout=2)

            response = self.async_run(run)
            rrs = response.get_rrset(
                response.answer, qname, dns.rdataclass.IN, dns.rdatatype.A
            )
            self.assertTrue(rrs is not None)
            seen = set([rdata.address for rdata in rrs])
            self.assertTrue("8.8.8.8" in seen)
            self.assertTrue("8.8.4.4" in seen)

    @unittest.skipIf(not _ssl_available, "SSL not available")
    def testQueryTLS(self):
        for address in query_addresses:
            qname = dns.name.from_text("dns.google.")

            async def run():
                q = dns.message.make_query(qname, dns.rdatatype.A)
                return await dns.asyncquery.tls(q, address, timeout=2)

            response = self.async_run(run)
            rrs = response.get_rrset(
                response.answer, qname, dns.rdataclass.IN, dns.rdatatype.A
            )
            self.assertTrue(rrs is not None)
            seen = set([rdata.address for rdata in rrs])
            self.assertTrue("8.8.8.8" in seen)
            self.assertTrue("8.8.4.4" in seen)

    @unittest.skipIf(not _ssl_available, "SSL not available")
    def testQueryTLSWithContext(self):
        for address in query_addresses:
            qname = dns.name.from_text("dns.google.")

            async def run():
                ssl_context = ssl.create_default_context()
                ssl_context.check_hostname = True
                q = dns.message.make_query(qname, dns.rdatatype.A)
                return await dns.asyncquery.tls(
                    q, address, timeout=2, ssl_context=ssl_context
                )

            response = self.async_run(run)
            rrs = response.get_rrset(
                response.answer, qname, dns.rdataclass.IN, dns.rdatatype.A
            )
            self.assertTrue(rrs is not None)
            seen = set([rdata.address for rdata in rrs])
            self.assertTrue("8.8.8.8" in seen)
            self.assertTrue("8.8.4.4" in seen)

    @unittest.skipIf(not _ssl_available, "SSL not available")
    def testQueryTLSWithSocket(self):
        for address in query_addresses:
            qname = dns.name.from_text("dns.google.")

            async def run():
                ssl_context = ssl.create_default_context()
                ssl_context.check_hostname = False
                async with await self.backend.make_socket(
                    dns.inet.af_for_address(address),
                    socket.SOCK_STREAM,
                    0,
                    None,
                    (address, 853),
                    2,
                    ssl_context,
                    None,
                ) as s:
                    # for basic coverage
                    await s.getsockname()
                    q = dns.message.make_query(qname, dns.rdatatype.A)
                    return await dns.asyncquery.tls(q, "8.8.8.8", sock=s, timeout=2)

            response = self.async_run(run)
            rrs = response.get_rrset(
                response.answer, qname, dns.rdataclass.IN, dns.rdatatype.A
            )
            self.assertTrue(rrs is not None)
            seen = set([rdata.address for rdata in rrs])
            self.assertTrue("8.8.8.8" in seen)
            self.assertTrue("8.8.4.4" in seen)

    def testQueryUDPFallback(self):
        for address in query_addresses:
            qname = dns.name.from_text(".")

            async def run():
                q = dns.message.make_query(qname, dns.rdatatype.DNSKEY)
                return await dns.asyncquery.udp_with_fallback(q, address, timeout=2)

            (_, tcp) = self.async_run(run)
            self.assertTrue(tcp)

    def testQueryUDPFallbackNoFallback(self):
        for address in query_addresses:
            qname = dns.name.from_text("dns.google.")

            async def run():
                q = dns.message.make_query(qname, dns.rdatatype.A)
                return await dns.asyncquery.udp_with_fallback(q, address, timeout=2)

            (_, tcp) = self.async_run(run)
            self.assertFalse(tcp)

    def testUDPReceiveQuery(self):
        async def run():
            async with await self.backend.make_socket(
                socket.AF_INET, socket.SOCK_DGRAM, source=("127.0.0.1", 0)
            ) as listener:
                listener_address = await listener.getsockname()
                async with await self.backend.make_socket(
                    socket.AF_INET, socket.SOCK_DGRAM, source=("127.0.0.1", 0)
                ) as sender:
                    sender_address = await sender.getsockname()
                    q = dns.message.make_query("dns.google", dns.rdatatype.A)
                    await dns.asyncquery.send_udp(sender, q, listener_address)
                    expiration = time.time() + 2
                    (_, _, recv_address) = await dns.asyncquery.receive_udp(
                        listener, expiration=expiration
                    )
                    return (sender_address, recv_address)

        (sender_address, recv_address) = self.async_run(run)
        self.assertEqual(sender_address, recv_address)

    def testUDPReceiveTimeout(self):
        async def arun():
            async with await self.backend.make_socket(
                socket.AF_INET, socket.SOCK_DGRAM, 0, ("127.0.0.1", 0)
            ) as s:
                try:
                    # for basic coverage
                    await s.getpeername()
                except Exception:
                    # we expect failure as we haven't connected the socket
                    pass
                await s.recvfrom(1000, 0.05)

        def run():
            self.async_run(arun)

        self.assertRaises(dns.exception.Timeout, run)

    @unittest.skipIf(not dns.query._have_httpx, "httpx not available")
    def testDOHGetRequest(self):
        async def run():
            nameserver_url = random.choice(KNOWN_ANYCAST_DOH_RESOLVER_URLS)
            q = dns.message.make_query("example.com.", dns.rdatatype.A)
            r = await dns.asyncquery.https(
                q, nameserver_url, post=False, timeout=4, family=family
            )
            self.assertTrue(q.is_response(r))

        self.async_run(run)

    @unittest.skipIf(not dns.query._have_httpx, "httpx not available")
    def testDOHPostRequest(self):
        async def run():
            nameserver_url = random.choice(KNOWN_ANYCAST_DOH_RESOLVER_URLS)
            q = dns.message.make_query("example.com.", dns.rdatatype.A)
            r = await dns.asyncquery.https(
                q, nameserver_url, post=True, timeout=4, family=family
            )
            self.assertTrue(q.is_response(r))

        self.async_run(run)

    @unittest.skipIf(not dns.query._have_httpx, "httpx not available")
    def testResolverDOH(self):
        async def run():
            res = dns.asyncresolver.Resolver(configure=False)
            res.nameservers = ["https://dns.google/dns-query"]
            answer = await res.resolve("dns.google", "A", backend=self.backend)
            seen = set([rdata.address for rdata in answer])
            self.assertTrue("8.8.8.8" in seen)
            self.assertTrue("8.8.4.4" in seen)

        self.async_run(run)

    @unittest.skipIf(not tests.util.have_ipv4(), "IPv4 not reachable")
    def testResolveAtAddress(self):
        async def run():
            answer = await dns.asyncresolver.resolve_at("8.8.8.8", "dns.google.", "A")
            seen = set([rdata.address for rdata in answer])
            self.assertIn("8.8.8.8", seen)
            self.assertIn("8.8.4.4", seen)

        self.async_run(run)

    @unittest.skipIf(not tests.util.have_ipv4(), "IPv4 not reachable")
    def testResolveAtName(self):
        async def run():
            answer = await dns.asyncresolver.resolve_at(
                "dns.google", "dns.google.", "A", family=socket.AF_INET
            )
            seen = set([rdata.address for rdata in answer])
            self.assertIn("8.8.8.8", seen)
            self.assertIn("8.8.4.4", seen)

        self.async_run(run)

    def testSleep(self):
        async def run():
            before = time.time()
            await self.backend.sleep(0.1)
            after = time.time()
            self.assertTrue(after - before >= 0.1)

        self.async_run(run)


@unittest.skipIf(not tests.util.is_internet_reachable(), "Internet not reachable")
class AsyncioOnlyTests(unittest.TestCase):
    def setUp(self):
        self.backend = dns.asyncbackend.set_default_backend("asyncio")

    def async_run(self, afunc):
        return asyncio.run(afunc())

    def testUseAfterTimeout(self):
        # Test #843 fix.
        async def run():
            qname = dns.name.from_text("dns.google")
            query = dns.message.make_query(qname, "A")
            sock = await self.backend.make_socket(socket.AF_INET, socket.SOCK_DGRAM)
            async with sock:
                # First do something that will definitely timeout.
                try:
                    response = await dns.asyncquery.udp(
                        query, "8.8.8.8", timeout=0.0001, sock=sock
                    )
                except dns.exception.Timeout:
                    pass
                except Exception:
                    self.assertTrue(False)
                # Now try to reuse the socket with a reasonable timeout.
                try:
                    response = await dns.asyncquery.udp(
                        query, "8.8.8.8", timeout=5, sock=sock
                    )
                    rrs = response.get_rrset(
                        response.answer, qname, dns.rdataclass.IN, dns.rdatatype.A
                    )
                    self.assertTrue(rrs is not None)
                    seen = set([rdata.address for rdata in rrs])
                    self.assertTrue("8.8.8.8" in seen)
                    self.assertTrue("8.8.4.4" in seen)
                except Exception:
                    self.assertTrue(False)

        self.async_run(run)


try:
    import sniffio
    import trio

    class TrioAsyncDetectionTests(AsyncDetectionTests):
        sniff_result = "trio"

        def async_run(self, afunc):
            return trio.run(afunc)

    class TrioNoSniffioAsyncDetectionTests(NoSniffioAsyncDetectionTests):
        expect_raise = True

        def async_run(self, afunc):
            return trio.run(afunc)

    class TrioAsyncTests(AsyncTests):
        def setUp(self):
            self.backend = dns.asyncbackend.set_default_backend("trio")

        def async_run(self, afunc):
            return trio.run(afunc)

except ImportError:
    pass


class MockSock:
    def __init__(self, wire1, from1, wire2, from2):
        self.family = socket.AF_INET
        self.first_time = True
        self.wire1 = wire1
        self.from1 = from1
        self.wire2 = wire2
        self.from2 = from2

    async def sendto(self, data, where, timeout):
        return len(data)

    async def recvfrom(self, bufsize, expiration):
        if self.first_time:
            self.first_time = False
            return self.wire1, self.from1
        else:
            return self.wire2, self.from2


class IgnoreErrors(unittest.TestCase):
    def setUp(self):
        self.q = dns.message.make_query("example.", "A")
        self.good_r = dns.message.make_response(self.q)
        self.good_r.set_rcode(dns.rcode.NXDOMAIN)
        self.good_r_wire = self.good_r.to_wire()
        dns.asyncbackend.set_default_backend("asyncio")

    def async_run(self, afunc):
        return asyncio.run(afunc())

    async def mock_receive(
        self,
        wire1,
        from1,
        wire2,
        from2,
        ignore_unexpected=True,
        ignore_errors=True,
        raise_on_truncation=False,
        good_r=None,
    ):
        if good_r is None:
            good_r = self.good_r
        s = MockSock(wire1, from1, wire2, from2)
        (r, when, _) = await dns.asyncquery.receive_udp(
            s,
            ("127.0.0.1", 53),
            time.time() + 2,
            ignore_unexpected=ignore_unexpected,
            ignore_errors=ignore_errors,
            raise_on_truncation=raise_on_truncation,
            query=self.q,
        )
        self.assertEqual(r, good_r)

    def test_good_mock(self):
        async def run():
            await self.mock_receive(self.good_r_wire, ("127.0.0.1", 53), None, None)

        self.async_run(run)

    def test_bad_address(self):
        async def run():
            await self.mock_receive(
                self.good_r_wire, ("127.0.0.2", 53), self.good_r_wire, ("127.0.0.1", 53)
            )

        self.async_run(run)

    def test_bad_address_not_ignored(self):
        async def abad():
            await self.mock_receive(
                self.good_r_wire,
                ("127.0.0.2", 53),
                self.good_r_wire,
                ("127.0.0.1", 53),
                ignore_unexpected=False,
            )

        def bad():
            self.async_run(abad)

        self.assertRaises(dns.query.UnexpectedSource, bad)

    def test_not_response_not_ignored_udp_level(self):
        async def abad():
            bad_r = dns.message.make_response(self.q)
            bad_r.id += 1
            bad_r_wire = bad_r.to_wire()
            s = MockSock(
                bad_r_wire, ("127.0.0.1", 53), self.good_r_wire, ("127.0.0.1", 53)
            )
            await dns.asyncquery.udp(self.good_r, "127.0.0.1", sock=s)

        def bad():
            self.async_run(abad)

        self.assertRaises(dns.query.BadResponse, bad)

    def test_bad_id(self):
        async def run():
            bad_r = dns.message.make_response(self.q)
            bad_r.id += 1
            bad_r_wire = bad_r.to_wire()
            await self.mock_receive(
                bad_r_wire, ("127.0.0.1", 53), self.good_r_wire, ("127.0.0.1", 53)
            )

        self.async_run(run)

    def test_bad_id_not_ignored(self):
        bad_r = dns.message.make_response(self.q)
        bad_r.id += 1
        bad_r_wire = bad_r.to_wire()

        async def abad():
            (r, wire) = await self.mock_receive(
                bad_r_wire,
                ("127.0.0.1", 53),
                self.good_r_wire,
                ("127.0.0.1", 53),
                ignore_errors=False,
            )

        def bad():
            self.async_run(abad)

        self.assertRaises(AssertionError, bad)

    def test_bad_wire(self):
        async def run():
            bad_r = dns.message.make_response(self.q)
            bad_r.id += 1
            bad_r_wire = bad_r.to_wire()
            await self.mock_receive(
                bad_r_wire[:10], ("127.0.0.1", 53), self.good_r_wire, ("127.0.0.1", 53)
            )

        self.async_run(run)

    def test_good_wire_with_truncation_flag_and_no_truncation_raise(self):
        async def run():
            tc_r = dns.message.make_response(self.q)
            tc_r.flags |= dns.flags.TC
            tc_r_wire = tc_r.to_wire()
            await self.mock_receive(
                tc_r_wire, ("127.0.0.1", 53), None, None, good_r=tc_r
            )

        self.async_run(run)

    def test_good_wire_with_truncation_flag_and_truncation_raise(self):
        async def agood():
            tc_r = dns.message.make_response(self.q)
            tc_r.flags |= dns.flags.TC
            tc_r_wire = tc_r.to_wire()
            await self.mock_receive(
                tc_r_wire, ("127.0.0.1", 53), None, None, raise_on_truncation=True
            )

        def good():
            self.async_run(agood)

        self.assertRaises(dns.message.Truncated, good)

    def test_wrong_id_wire_with_truncation_flag_and_no_truncation_raise(self):
        async def run():
            bad_r = dns.message.make_response(self.q)
            bad_r.id += 1
            bad_r.flags |= dns.flags.TC
            bad_r_wire = bad_r.to_wire()
            await self.mock_receive(
                bad_r_wire, ("127.0.0.1", 53), self.good_r_wire, ("127.0.0.1", 53)
            )

        self.async_run(run)

    def test_wrong_id_wire_with_truncation_flag_and_truncation_raise(self):
        async def run():
            bad_r = dns.message.make_response(self.q)
            bad_r.id += 1
            bad_r.flags |= dns.flags.TC
            bad_r_wire = bad_r.to_wire()
            await self.mock_receive(
                bad_r_wire,
                ("127.0.0.1", 53),
                self.good_r_wire,
                ("127.0.0.1", 53),
                raise_on_truncation=True,
            )

        self.async_run(run)

    def test_bad_wire_not_ignored(self):
        bad_r = dns.message.make_response(self.q)
        bad_r.id += 1
        bad_r_wire = bad_r.to_wire()

        async def abad():
            await self.mock_receive(
                bad_r_wire[:10],
                ("127.0.0.1", 53),
                self.good_r_wire,
                ("127.0.0.1", 53),
                ignore_errors=False,
            )

        def bad():
            self.async_run(abad)

        self.assertRaises(dns.message.ShortHeader, bad)

    def test_trailing_wire(self):
        async def run():
            wire = self.good_r_wire + b"abcd"
            await self.mock_receive(
                wire, ("127.0.0.1", 53), self.good_r_wire, ("127.0.0.1", 53)
            )

        self.async_run(run)

    def test_trailing_wire_not_ignored(self):
        wire = self.good_r_wire + b"abcd"

        async def abad():
            await self.mock_receive(
                wire,
                ("127.0.0.1", 53),
                self.good_r_wire,
                ("127.0.0.1", 53),
                ignore_errors=False,
            )

        def bad():
            self.async_run(abad)

        self.assertRaises(dns.message.TrailingJunk, bad)



import dns.rdata
import dns.rdataset
import dns.rdtypes.IN.SRV


def test_processing_order_shuffle():
    rds = dns.rdataset.from_text("in", "a", 300, "10.0.0.1", "10.0.0.2", "10.0.0.3")
    seen = set()
    for i in range(100):
        po = rds.processing_order()
        assert len(po) == 3
        for j in range(3):
            assert rds[j] in po
        seen.add(tuple(po))
    assert len(seen) == 6


def test_processing_order_priority_mx():
    rds = dns.rdataset.from_text("in", "mx", 300, "10 a", "20 b", "20 c")
    seen = set()
    for i in range(100):
        po = rds.processing_order()
        assert len(po) == 3
        for j in range(3):
            assert rds[j] in po
        assert rds[0] == po[0]
        seen.add(tuple(po))
    assert len(seen) == 2


def test_processing_order_priority_weighted():
    rds = dns.rdataset.from_text(
        "in", "srv", 300, "1 10 1234 a", "2 90 1234 b", "2 10 1234 c"
    )
    seen = set()
    weight_90_count = 0
    weight_10_count = 0
    for i in range(100):
        po = rds.processing_order()
        assert len(po) == 3
        for j in range(3):
            assert rds[j] in po
        assert rds[0] == po[0]
        assert isinstance(po[1], dns.rdtypes.IN.SRV.SRV)
        if po[1].weight == 90:
            weight_90_count += 1
        else:
            assert po[1].weight == 10
            weight_10_count += 1
        seen.add(tuple(po))
    assert len(seen) == 2
    # We can't assert anything with certainty given these are random
    # draws, but it's super likely that weight_90_count > weight_10_count,
    # so we just assert that.
    assert weight_90_count > weight_10_count


def test_processing_order_priority_naptr():
    rds = dns.rdataset.from_text(
        "in",
        "naptr",
        300,
        "1 10 a b c foo.",
        "1 20 a b c foo.",
        "2 10 a b c foo.",
        "2 10 d e f bar.",
    )
    seen = set()
    for i in range(100):
        po = rds.processing_order()
        assert len(po) == 4
        for j in range(4):
            assert rds[j] in po
        assert rds[0] == po[0]
        assert rds[1] == po[1]
        seen.add(tuple(po))
    assert len(seen) == 2


def test_processing_order_empty():
    rds = dns.rdataset.from_text("in", "naptr", 300)
    po = rds.processing_order()
    assert po == []


def test_processing_singleton_priority():
    rds = dns.rdataset.from_text("in", "mx", 300, "10 a")
    po = rds.processing_order()
    assert po == [rds[0]]


def test_processing_singleton_weighted():
    rds = dns.rdataset.from_text("in", "srv", 300, "1 10 1234 a")
    po = rds.processing_order()
    assert po == [rds[0]]


def test_processing_all_zero_weight_srv():
    rds = dns.rdataset.from_text(
        "in", "srv", 300, "1 0 1234 a", "1 0 1234 b", "1 0 1234 c"
    )
    seen = set()
    for i in range(100):
        po = rds.processing_order()
        assert len(po) == 3
        for j in range(3):
            assert rds[j] in po
        seen.add(tuple(po))
    assert len(seen) == 6


def test_processing_order_uri():
    # We're testing here just to provide coverage for URI methods; the
    # testing of the weighting algorithm is done above in tests with
    # SRV.
    rds = dns.rdataset.from_text(
        "in",
        "uri",
        300,
        '1 1 "ftp://ftp1.example.com/public"',
        '2 2 "ftp://ftp2.example.com/public"',
        '3 3 "ftp://ftp3.example.com/public"',
    )
    po = rds.processing_order()
    assert len(po) == 3
    for i in range(3):
        assert po[i] == rds[i]


def test_processing_order_svcb():
    # We're testing here just to provide coverage for SVCB methods; the
    # testing of the priority algorithm is done above in tests with
    # MX and NAPTR.
    rds = dns.rdataset.from_text(
        "in",
        "svcb",
        300,
        "1 . mandatory=alpn alpn=h2",
        "2 . mandatory=alpn alpn=h2",
        "3 . mandatory=alpn alpn=h2",
    )
    po = rds.processing_order()
    assert len(po) == 3
    for i in range(3):
        assert po[i] == rds[i]
